#
# $Id: alcGObjects.py 862 2007-11-05 23:07:02Z trylon $
#
#    Copyright (C) 2005-2006  Alcugs pyprp Project Team
#    See the file AUTHORS for more info about the team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Please see the file COPYING for the full license.
#    Please see the file DISCLAIMER for more details, before doing nothing.
#

# Help library

try:
    import Blender
    try:
        from Blender import Mesh
    except Exception, detail:
        print detail
except ImportError:
    pass

import random, md5,math, binascii, struct

def distance(p1,p2):
    a1=p1[0]
    a2=p2[0]
    b1=p1[1]
    b2=p2[1]
    c1=p1[2]
    c2=p2[2]
    return math.sqrt(pow(a1-a2,2) + pow(b1-b2,2) + pow(c1-c2,2))


def addProp(object,name,val):
    try:
        p=object.getProperty(name)
        p.setData(val)
    except (AttributeError, RuntimeError):
        object.addProperty(name,val)


def getProp(object,property):
    try:
        p=object.getProperty(property)
        return(p.getData())
    except (AttributeError, RuntimeError):
        return None


def getMatrix(object):
    try:
        m=Blender.Mathutils.Matrix(object.getMatrix())
    except TypeError:
        i=object.getMatrix()
        m=Blender.Mathutils.Matrix([i[0][0],i[0][1],i[0][2],i[0][3]],
                                   [i[1][0],i[1][1],i[1][2],i[1][3]],
                                   [i[2][0],i[2][1],i[2][2],i[2][3]],
                                   [i[3][0],i[3][1],i[3][2],i[3][3]])
    return m


def alcHex2Ascii(what,size=4):
    if size==4:
        type=">I"
    elif size==2:
        type=">H"
    else:
        type=">B"
    a=struct.pack(type,what)
    return(binascii.hexlify(a).upper())


def alcAscii2Hex(name,size=4):
    if size==4:
        type=">I"
    elif size==2:
        type=">H"
    else:
        type=">B"

    while len(name)<size*2:
        name = "0" + name
    b=binascii.unhexlify(str(name))
    r, = struct.unpack(type,b)
    return(r)


def alcUniqueName(name,seed1=0,seed2=0,key="x"):
    if seed1==0 or seed2==0:
        w = str(name) + "%i" %int(random.random()*10000)
        rid=md5.new(w).hexdigest().upper()
        mname = str(name)[:10] + "%s%s" %(key[:1],rid[:6])
        return mname
    else:
        rid=((int(random.random()*10000)*100+setidx)*10 + i)
        mname = name[:10] + "%s%06X" %(key[:1],rid)
        return mname


def alcCreateLinkInPoint(name="LinkInPointDefault",where=None,page=0):
    obj = Blender.Object.New('Empty',name)
    scene = Blender.Scene.GetCurrent()
    scene.link(obj)
    #obj.addProperty("name",name)
    obj.addProperty("alctype","swpoint")
    if page!=0:
        obj.addProperty("page_num",str(page))
    if where==None:
        obj.setLocation(Blender.Window.GetCursorPos())
    else:
        matrix=where.get()
        matrix.transpose()
        obj.setMatrix(matrix)
    obj.layers = [2,]
    try:
        obj.setDrawMode(9)
    except:
        obj.setDrawMode(10)
    obj.setDrawType(2)
    #Blender.Redraw()
    return obj


def alcCreateRegion(name="Region",where=None,page=0):
    #tempname = alcUniqueName(name,0,0,'X')
    rgnsize = 4
    rgnoffset = 0 - (rgnsize/2)
    obj = alcCreateBox(name,rgnsize,rgnoffset,rgnoffset,rgnoffset)
    obj.addProperty("alctype","region")
    obj.addProperty("prpregion","unknown")

    if page!=0:
        obj.addProperty("page_num",str(page))
    if where==None:
        obj.setLocation(Blender.Window.GetCursorPos())
    else:
        matrix=where.get()
        matrix.transpose()
        obj.setMatrix(matrix)
    obj.layers = [2,]
    obj.drawType = 2
    #Blender.Redraw()
    return obj


def alcCreatePanicLnkRegion(name="PanicLnk",where=None,page=0):
    obj = alcCreateRegion(name,where,page)
    p = obj.getProperty("prpregion")
    p.setData("paniclnkrgn")
    return obj

def alcCreateFootstepRegion(name="FootStep",where=None,page=0):
    obj = alcCreateRegion(name,where,page)
    p = obj.getProperty("prpregion")
    p.setData("footsteprgn")
    obj.addProperty("footstepsound",int(0))
    return obj

def alcCreateClickableRegion(where=None,page=0):
    obj = alcCreateRegion("ClickableRegion",where,page)
    p = obj.getProperty("prpregion")
    p.setData("clickregion")
    return obj

def alcCreateCameraRegion(name="CameraRgn",where=None,page=0):
    obj = alcCreateRegion(name,where,page)
    p = obj.getProperty("prpregion")
    p.setData("camerargn")
    obj.addProperty("camera","")
    obj.addProperty("setDefCam","false")
    return obj

def alcCreateSwimRegion(name="SwimDetRgn",where=None,page=0):
    obj = alcCreateRegion(name,where,page)
    p = obj.getProperty("prpregion")
    p.setData("swimrgn")
    return obj

def alcCreateSwimSurface(name="SwimSfc",where=None,page=0):
    tempname = alcUniqueName(name,0,0,'sfc')
    rgnsize = 4
    rgnoffset = 0 - (rgnsize/2)
    obj = alcCreatePlane(tempname,rgnsize,rgnoffset,rgnoffset,0)
    scene = Blender.Scene.GetCurrent()
    #scene.link(obj)
    obj.addProperty("alctype","region")
    obj.addProperty("prpregion","swimplainsfc")

    if page!=0:
        obj.addProperty("page_num",str(page))
    if where==None:
        obj.setLocation(Blender.Window.GetCursorPos())
    else:
        matrix=where.get()
        matrix.transpose()
        obj.setMatrix(matrix)
    obj.layers = [2,]
    obj.drawType = 2
    #Blender.Redraw()
    return obj


def alcCreateMesh(name,vertices,faces):
    obj = Mesh.New(name)
    obj.verts.extend(vertices)
    obj.faces.extend(faces)
    scene = Blender.Scene.GetCurrent()
    sobj=scene.objects.new(obj,name)
    sobj.select(False)
    sobj.setName(name)
    return sobj


def alcCreateBox(name,size,x=0,y=0,z=0,bReverse=False):
    obj = Mesh.New(name)
    coords = [ [x,y,z], [size+x,y,z], [x,size+y,z], [x,y,size+z], [size+x,size+y,size+z], [size+x,size+y,z], [size+x,y,size+z], [x,size+y,size+z] ]
    obj.verts.extend(coords)
    if bReverse:
        faces = [ [0,1,5,2], [5,4,7,2], [1,6,4,5], [0,3,6,1], [3,7,4,6], [0,2,7,3] ]
    else:
        faces = [ [0,2,5,1], [5,2,7,4], [1,5,4,6], [0,1,6,3], [3,6,4,7], [0,3,7,2] ]
    obj.faces.extend(faces)
    scene = Blender.Scene.GetCurrent()
    sobj = scene.objects.new(obj,name)
    sobj.select(False)
    return sobj


def alcCreatePlane(name,size,x=0,y=0,z=0):
    obj = Mesh.New(name)
    coords = [ [x,y,z], [size+x,y,z], [x,size+y,z], [size+x,size+y,z] ]
    obj.verts.extend(coords)
    face = [0,1,3,2]
    obj.faces.extend(face)
    scene = Blender.Scene.GetCurrent()
    sobj = scene.objects.new(obj,name)
    sobj.select(False)
    return sobj


def alcFindBlenderText(name):
    try:
        out=Blender.Text.Get(name)
    except NameError:
        out=Blender.Text.New(name)
    return out
    

def alcCreateClimbRegion(name="ClimbRgn",where=None,page=0):
    obj = alcCreateRegion("ClickableRegion",where,page)
    p = obj.getProperty("prpregion")
    p.setData("climbregion")

    obj.addProperty("bottomFlag","00")
    obj.addProperty("climbOffset",int(0))
    obj.addProperty("climbHeight",int(0))
    return obj
    
def deldefaultproperty(obj,propertyname,defaultvalue):
    try:
        p=obj.getProperty(propertyname)
        if(p.getData() == defaultvalue):
            obj.removeProperty(p)
    except (AttributeError, RuntimeError):
        print "Error removing %s property" % propertyname
        
def getFloatPropertyOrDefault(obj,propertyname,defaultvalue):
    try:
        p=obj.getProperty(propertyname)
        if(p.type == "FLOAT"):
            var=float(p.getData())
        else:
            var=defaultvalue
    except (AttributeError, RuntimeError):
        var=defaultvalue
    return var

def getHexPropertyOrDefault(obj,propertyname,bytesize,defaultvalue):
    try:
        p=obj.getProperty(propertyname)
        if(p.type == "STRING"):
            var=alcAscii2Hex(str(p.getData()),bytesize)
        else:
            var=defaultvalue
    except (AttributeError, RuntimeError):
        var=defaultvalue
        pass
    return var

def getTextPropertyOrDefault(obj,propertyname,defaultvalue):
    try:
        p=obj.getProperty(propertyname)
        if(p.type == "STRING"):
            var=str(p.getData())
        else:
            var=defaultvalue
    except (AttributeError, RuntimeError):
        var=defaultvalue
        pass
    return var

def getStrIntPropertyOrDefault(obj,propertyname,defaultvalue):
    try:
        p=obj.getProperty(propertyname)
        if(p.type == "STRING"):
            var=int(p.getData())
        else:
            var=defaultvalue
    except (AttributeError, RuntimeError):
        var=defaultvalue
        pass
    return var

def getIntPropertyOrDefault(obj,propertyname,defaultvalue):
    try:
        p=obj.getProperty(propertyname)
        if(p.type == "INT"):
            var=int(p.getData())
        else:
            var=defaultvalue
    except (AttributeError, RuntimeError):
        var=defaultvalue
        pass
    return var

def getBoolPropertyOrDefault(obj,propertyname,defaultvalue):
    try:
        p=obj.getProperty(propertyname)
        if(p.type == "BOOL"):
            var=int(p.getData() > 0)
        else:
            var=int(defaultvalue > 0)
    except (AttributeError, RuntimeError):
        var=int(defaultvalue > 0)
        pass
    return var

def alcCreateSoftVolumePlane():
    tempname = alcUniqueName("SoftVol",0,0,'plane')
    rgnsize = 4
    rgnoffset = 0 - (rgnsize/2)
    obj = alcCreatePlane(tempname,rgnsize,rgnoffset,rgnoffset,0)
    scene = Blender.Scene.GetCurrent()
    obj.addProperty("alctype","svconvex")
    obj.setLocation(Blender.Window.GetCursorPos())
    obj.layers = [6,]
    obj.drawType = 2
    return obj


def alcCreateSoftVolumeCube():
    tempname = alcUniqueName("SoftVol",0,0,'cube')
    rgnsize = 4
    rgnoffset = 0 - (rgnsize/2)
    obj = alcCreateBox(tempname,rgnsize,rgnoffset,rgnoffset,rgnoffset,True)
    obj.addProperty("alctype","svconvex")
    obj.setLocation(Blender.Window.GetCursorPos())
    obj.layers = [6,]
    obj.drawType = 2
    return obj
