#
# $Id: alc_AbsClasses.py 455 2006-04-21 20:18:42Z AdamJohnso $
#
#    Copyright (C) 2005-2006  Alcugs pyprp Project Team
#    See the file AUTHORS for more info about the team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Please see the file COPYING for the full license.
#    Please see the file DISCLAIMER for more details, before doing nothing.
#

try:
    import Blender
except ImportError:
    pass

import math, struct
from enum import Enum
from alchexdump import *
from alcurutypes import *
import alcconfig, alchexdump
from alc_Messages import *
from alc_Creatable import *
#from alc_Classes import *
from alcurugeom import *

class plSynchedObject(hsKeyedObject):                       #Type 0x28
    def __init__(self,parent,name="unnamed",type=0x28):
        hsKeyedObject.__init__(self,parent,name,type)
        self.Flags=0
        self.Count=0
        self.SDLLinks=[]
    

    def read(self,buf):
        hsKeyedObject.read(self,buf)
        self.Flags = buf.Read32()
        #if not self.Flags in (0x00,0x04,0x10,0x20,0x28,0x38,0x80):
        #    raise RuntimeError, "plSynchedObject contains unknown Flag %04X" % self.Flags
        if (self.getVersion()==5 and self.Flags & 0x10) or (self.getVersion()==6 and not self.Flags & 0x04): #We need to document these flags
            self.Count = buf.Read16()
            for i in range(self.Count):
                #s = buf.ReadSafeString(self.getVersion())
                s = Ustr()
                s.setType(self.getVersion())
                s.read(buf)
                self.SDLLinks.append(s)
                #if not str(s) in ("Physical","AGMaster","Responder","Clothing","Avatar","Layer","Sound","XRegion","MorphSequence","ParticleSystem"):
                #    raise RuntimeError, "Unknown %s sdl link in sceneobject" % str(s)


    def write(self,buf):
        hsKeyedObject.write(self,buf)
        if len(self.SDLLinks)!=0:
            self.Flags |= 0x10
        buf.Write32(self.Flags)
        if self.Flags & 0x10:
            buf.Write16(len(self.SDLLinks))
            for s in self.SDLLinks:
                # Commented out WriteSafeString; use Ustr.write instead
                #s.WriteSafeString(s, self.getVersion())
                s.setType(self.getVersion())
                s.write(buf,1)


    def changePageRaw(self,sid,did,stype,dtype):
        hsKeyedObject.changePageRaw(self,sid,did,stype,dtype)

    def dump(self):
        hsKeyedObject.dump(self)
        print "Synchronized Object:"
        print "Flag %08X" % self.Flags
        print "SDL Links: ",
        for i in self.SDLLinks:
            print str(i),


class plObjInterface(plSynchedObject):                      #Type 0x10
    def __init__(self,parent,name="unnamed",type=None):
        plSynchedObject.__init__(self,parent,name,type)
        self.parentref=UruObjectRef(self.getVersion())
        self.BitFlags=hsBitVector()

    def read(self,buf):
        plSynchedObject.read(self,buf)
        self.parentref.read(buf)
        self.BitFlags.read(buf)


    def write(self,buf):
        plSynchedObject.write(self,buf)
        self.parentref.write(buf)
        self.BitFlags.write(buf)

    def changePageRaw(self,sid,did,stype,dtype):
        plSynchedObject.changePageRaw(self,sid,did,stype,dtype)
        self.parentref.changePageRaw(sid,did,stype,dtype)

    def dump(self):
        plSynchedObject.dump(self)
        print "Parent Object: %s" % self.parentref


class plAudible(hsKeyedObject):                             #Type 0x12
    def __init__(self,parent,name="unnamed",type=None):
        hsKeyedObject.__init__(self,parent,name,type)


class plModifier(plSynchedObject):                          #Type 0x1E
    def __init__(self,parent,name="unnamed",type=None):
        plSynchedObject.__init__(self,parent,name,type)


class plSingleModifier(plModifier):                         #Type 0x1F
    def __init__(self,parent,name="unnamed",type=None):
        plModifier.__init__(self,parent,name,type)
        self.bitVector = hsBitVector()


    def read(self,buf):
        plModifier.read(self,buf)
        self.bitVector.read(buf)
    
    
    def write(self,buf):
        plModifier.write(self,buf)
        self.bitVector.write(buf)


class plDetectorModifier(plSingleModifier):                 #Type 0x24
    def __init__(self,parent,name="unnamed",type=None):
        plSingleModifier.__init__(self,parent,name,type)
        #Format
        if (self.getVersion()==6):
            self.logic = hsTArray([0x2C],6)
        else:
            self.logic = hsTArray([0x2D,0xA6],5,True)
        self.fRemoteMod = UruObjectRef(self.getVersion())
        self.fProxyKey = UruObjectRef(self.getVersion())
    
    
    def read(self,buf):
        plSingleModifier.read(self,buf)
        self.logic.ReadVector(buf)
        self.fRemoteMod.read(buf)
        self.fProxyKey.read(buf)
    
    
    def write(self,buf):
        plSingleModifier.write(self,buf)
        self.logic.update(self.Key)
        self.logic.WriteVector(buf)
        self.fRemoteMod.write(buf)
        self.fProxyKey.write(buf)

    
    def dump(self):
        plSingleModifier.dump(self)
        print "Detector Modifier"
        print self.logic.size + " LogicModifer refs in hsTArray"
        print "ObjRef2: " + self.fProxyKey.Key.name


    def changePageRaw(self,sid,did,stype,dtype):
        plSingleModifier.changePageRaw(self,sid,did,stype,dtype)
        self.logic.changePageRaw(sid,did,stype,dtype)
        self.fRemoteMod.changePageRaw(sid,did,stype,dtype)
        self.fProxyKey.changePageRaw(sid,did,stype,dtype)


class plMultiModifier(plModifier):                          #Type 0x27
    def __init__(self,parent,name="unnamed",type=None):
        plModifier.__init__(self,parent,name,type)
        self.BitVector=hsBitVector()


    def read(self,buf):
        plModifier.read(self,buf)
        self.BitVector.read(buf)


    def write(self,buf):
        plModifier.write(self,buf)
        self.BitVector.write(buf)


class plCollisionDetector(plDetectorModifier):              #Type 0x2C
    Type = \
    { \
        "kTypeEnter"    :  0x1, \
        "kTypeExit"     :  0x2, \
        "kTypeAny"      :  0x4, \
        "kTypeUnEnter"  :  0x8, \
        "kTypeUnExit"   : 0x10, \
        "kTypeBump"     : 0x20  \
    }

    def __init__(self,parent,name="unnamed",type=None):
        plDetectorModifier.__init__(self,parent,name,type)
        self.fType = plCollisionDetector.Type["kTypeAny"] # default to any

    def read(self,buf):
        plDetectorModifier.read(self,buf)
        self.fType = buf.ReadByte()


    def write(self,buf):
        plDetectorModifier.write(self,buf)
        buf.WriteByte(self.fType)


class plConditionalObject(hsKeyedObject):                   #Type 0x2E

    Flags = \
    { \
        "kLocalElement" : 0, \
        "kNOT"          : 1  \
    }
    
    def __init__(self,parent,name="unnamed",type=None):
        hsKeyedObject.__init__(self,parent,name,type)
        self.bSatisfied = False #BYTE - Was self.i10
        self.fToggle = False #byte - Was self.i14
    
    
    def read(self,buf):
        hsKeyedObject.read(self,buf)
        self.bSatisfied = buf.ReadByte()
        self.fToggle = buf.ReadByte()
    
    
    def write(self,buf):
        hsKeyedObject.write(self,buf)
        buf.WriteByte(self.bSatisfied)
        buf.WriteByte(self.fToggle)
    
    
    def dump(self):
        print "Conditional Object"
        print "bSatisfied: " + self.bSatisfied
        print "fToggle: " + self.fToggle



class plLightInfo(plObjInterface):                          #Type 0x54 (Uru)
    
    kDisable = 0		
    kLPObsolete = 1 		
    kLPCastShadows = 2
    kLPMovable = 3
    kLPHasIncludes = 4
    kLPIncludesChars = 5
    kLP_OBSOLECTE_0 = 6
    kLPOverAll = 7
    kLPHasSpecular = 8
    kLPShadowOnly = 9
    kLPShadowLightGroup = 10
    kLPForceProj = 11
    kNumProps = 12

    def __init__(self,parent,name="unnamed",type=None):
        plObjInterface.__init__(self,parent,name,type)
        try: #Quick, dirty fix for NameError bug with classes from alcurugeom
            self.ambient = RGBA('1.0','1.0','1.0','1.0',type=1)
        except NameError:
            #print "Damnit! Need reimport alcurugeom.py"
            from alcurugeom import RGBA,hsMatrix44
            self.ambient = RGBA('1.0','1.0','1.0','1.0',type=1)

        self.diffuse = RGBA()
        self.specular = RGBA()
        
        self.LightToLocal = hsMatrix44()
        self.LocalToLight = hsMatrix44()
        self.LightToWorld = hsMatrix44()
        self.WorldToLight = hsMatrix44()
        
        self.layerint = UruObjectRef(self.getVersion()) #plLayerInterface
        self.softvol = UruObjectRef(self.getVersion()) #plSoftVolume
        self.scenenode = UruObjectRef(self.getVersion()) #Dunno
        
        self.visregs = hsTArray([],self.getVersion()) #plVisRegion[]
    
    def read(self,buf):
        plObjInterface.read(self,buf)
        self.ambient.read(buf)
        self.diffuse.read(buf)
        self.specular.read(buf)
        self.LightToLocal.read(buf)
        self.LocalToLight.read(buf)
        self.LightToWorld.read(buf)
        self.WorldToLight.read(buf)
        self.layerint.read(buf)
        self.softvol.read(buf)
        self.scenenode.read(buf)


        self.visregs.read(buf)
    
    def write(self,buf):
        plObjInterface.write(self,buf)
        self.ambient.write(buf)
        self.diffuse.write(buf)
        self.specular.write(buf)
        self.LightToLocal.write(buf)
        self.LocalToLight.write(buf)
        self.LightToWorld.write(buf)
        self.WorldToLight.write(buf)
        self.layerint.write(buf)
        self.softvol.write(buf)
        self.scenenode.write(buf)

        self.visregs.write(buf)
    
    def changePageRaw(self,sid,did,stype,dtype):
        plObjInterface.changePageRaw(self,sid,did,stype,dtype)
        self.softvol.changePageRaw(sid,did,stype,dtype)
        self.layerint.changePageRaw(sid,did,stype,dtype)
        self.scenenode.changePageRaw(sid,did,stype,dtype)
        self.visregs.changePageRaw(sid,did,stype,dtype)


class plLogicModBase(plSingleModifier):                   #Type 0x4F
    Flags = \
    { \
        "kLocalElement"      : 0, \
        "kReset"             : 1, \
        "kTriggered"         : 2, \
        "kOneShot"           : 3, \
        "kRequestingTrigger" : 4, \
        "kTypeActivator"     : 5, \
        "kMultiTrigger"      : 6  \
    }
    def __init__(self,parent,name="unnamed",type=None):
        plSingleModifier.__init__(self,parent,name,type)
        self.MsgCount = 0x00000000
        self.Messages = [] # plMessages
        self.Notify = PrpMessage(0x02E8,self.getVersion())
        self.fFlags = hsBitVector()
        self.fDisabled = 0x00

    def read(self,buf):
        plSingleModifier.read(self,buf)
        self.MsgCount = buf.Read32()
        assert(self.MsgCount == 0)
        localType = buf.Read16()
        assert(localType == 0x02E8) # plNotifyMsg type
        self.Notify.read(buf)
        
        self.fFlags.read(buf);
        
        self.fDisabled = buf.ReadByte()

    def write(self,buf):
        plSingleModifier.write(self,buf)
        buf.Write32(self.MsgCount)
        buf.Write16(self.Notify.msgtype)
        self.Notify.update(self.Key)
        self.Notify.write(buf)
        self.fFlags.write(buf)
        buf.WriteByte(self.fDisabled)


class plAGAnim(plSynchedObject):                #Type 0x6B
    def __init(self,parent,name="unnamed",type=None):
        plSynchedObject.__init__(self,parent,name,type)
        self.AnimName = 0
        self.StartTime = 0
        self.EndTime = 0
        self.AGApp = []
    
    def read(self,buf):
        plSynchedObject.read(self,buf)
        self.AnimName = buf.ReadSafeString()
        self.StartTime = buf.ReadFloat()
        self.StopTime = buf.ReadFloat()
        i = buf.Read32()
        for j in range(i):
            self.AGApp[j] = plAGApplicator()
            self.AGApp[j].read(buf)
    
    def write(self,buf):
        plSynchedObject.write(self,buf)
        buf.WriteSafeString(self.AnimName)
        buf.WriteFloat(self.StartTime)
        buf.WriteFloat(self.StopTime)
        buf.Write32(len(self.AGApp))
        for i in range(len(self.AGApp)):
            self.AGApp[i].write(buf)
    
    def dump(self):
        print "Animation: " + self.AnimName


class plRegionBase(plObjInterface):             #Type 0x0118
    def __init__(self,parent,name="unnamed",type=None):
        plObjInterface.__init__(self,parent,name,type)