#
# $Id: alc_AdvScript.py 777 2007-11-2 9:00:00Z Trylon $
#
#    Copyright (C) 2005-2006  Alcugs pyprp Project Team
#    See the file AUTHORS for more info about the team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite ##0, Boston, MA  02111-1#07  USA
#
#    Please see the file COPYING for the full license.
#    Please see the file DISCLAIMER for more details, before doing nothing.

#    This file contains the basic classes for interpreting the advanced settings of objects
#    It uses yaml for the script

import sys, os
import yaml

try:
    import Blender
except ImportError:
    pass

from alcGObjects import *


# Guide for PYPRP programmers:
# 
# This set of classes reads in a yaml (www.yaml.org) formatted scriptset
#
#
#
#
# The AlcScript class reads in the data provided in the init, and keeps a list of root Scriptobjects
# You can obtain a root script object by using the function <AlcScript>.FindObject(name[,type])
# 
# Each ScriptObject has a list ob subobjects
# There are two kinds of subobjects:
# - Objects that have subobjects
#   These are callable with object.subobject, and you can use object.Contains('subobject') to check for it
#   
# - Objects with a only value
#   These are callable with parentobject['key']
#   This also includes 
#
# Ideally they should be used in the export_obj functions of the various plasma Objects.
#
# Example code for a logicmodifier
        
    

class AlcScriptNode:
    def __init__(self,key,data,level=0):
        self.children = []
        self.key = key.lower()
        if AlcScript.Debug:
            self.value="<Node '" + self.key + "'>"    
        self.HasNamelessChildren = False

        
        self.hasValue = False

        if(type(data) == dict):
            if AlcScript.Debug:
                for i in range(level):
                    print " ",
                print "<" + str(self.key) + ">"
            # it has more children
            for key in data.keys():
                node = AlcScriptNode(key,data[key],level+1)
                self.children.append(node)
            if AlcScript.Debug:
                for i in range(level):
                    print " ",
                print "</" + str(self.key) + ">"
                
        elif type(data) == list and len(data) > 0 and type(data[0] == dict):
            # if it's a list, e.g. for python args, 
            # self.data will become a list of ScriptObjects
            # then make a base Node for each list object, and add 
            # them to the list
            # if it's read in, the value will be a list of AlcScriptNodes
            
            index = 0
            self.value = [] # 
            if AlcScript.Debug:
                for i in range(level):
                    print " ",
                print self.key, "="
                for i in range(level+1):
                    print " ",
                print "["
            
            for obj in data:
                node = AlcScriptNode("Node",obj,level+2)
                self.value.append(node)
                index += 1
                self.HasNamelessChildren = True
                if AlcScript.Debug:
                    for i in range(level+1):
                        print " ",
                    print ","

            if AlcScript.Debug:
                for i in range(level+1):
                    print " ",
                print "]"
                
        else:
            self.value = data
            self.hasValue = True
            if AlcScript.Debug:
                for i in range(level):
                    print " ",
                print self.key, "=" , self.value.__repr__()
        
    def HasChildren(self):
        if(len(self.children) > 0):
            return True
        else:
            return False
    
    def Contains(self,key):
        for child in self.children:
            if(child.key == key and child.HasChildren()):
               return True
        return False
    
    def _Children(self,key):
        childlist = []
        for child in self.children:
            if(child.key == key and child.HasChildren()):
               childlist.append(child)
        
        if len(childlist) == 0:
            return None
        elif len(childlist) == 1:
            return childlist[0]
        else:
            return childlist

    def __getattr__(self, name):
        return self._Children(name.lower())

    def _Attr(self,attr):
        for child in self.children:
            if(child.key == attr):
                if not child.HasChildren():
                    return child.value
        return None

    def _AttrCount(self):
        count = 0;
        for child in self.children:
            if(not child.HasChildren()):
                return child.value
        return count

    def __len__(self):
        return self._AttrCount()
  
    def __getitem__(self, key):
        item = self._Attr(key.lower())
        if(item != None):
            return item
        else:
            return ""
        
    def __contains__(self, item):
        if(self._Attr(item.lower()) != None):
            return True
        else:
            return False
            
    def __repr__(self,level=0):
        str = ""
        if level==0:
           str += "\n" 
        if(self.HasChildren()):
            for i in range(level):
                str += " "
            str += "<AlcNode " + self.key.__repr__() + ">\n"
            
            for child in self.children:
                str += child.__repr__(level+1)
                
            for i in range(level):
                str += " "
            str += "</AlcNode>\n"
        else:
            for i in range(level):
                str += " "
            str += self.key + "=" + self.value.__repr__() + " (" + type(self.value).__repr__() + ")\n"
        return str
        
    def __str__(self):
        return str(self.value)
        

    # Not really a not-equal check, you should only use this to check if it's not equal to "None"
    def __ne__(self,obj):
        if(type(obj) == bool):
            return (not obj)
        elif(type(obj) != type(self)):
            return True
        else:
            return True

    def __eq__(self,obj):
        if(type(obj) == bool):
            return (obj)
        elif(type(obj) != type(self)):
            return False
        else:
            return False
           
        
class AlcScript:
    current=None

    Debug=True
    def __init__(self,text,level=0):
        self.children = []
        structure = yaml.safe_load(text)
        print text
        print structure
        
        if type(structure) == dict:
            for key in structure.keys():
                node = AlcScriptNode(key,structure[key],level)
                self.children.append(node)
                if AlcScript.Debug:
                    print "--"
        elif type(structure) == list:
            for data in structure:
                for key in data.keys():
                    node = AlcScriptNode(key,data[key],level)
                    self.children.append(node)
                    if AlcScript.Debug:
                        print "--"
       
    def FindNamedObject(self,name,type=''):
        for child in self.children:
            if child.key == name and child['type'] == type:
                return child
        return None

    def FindTypedObject(self,name,type=''):
        for child in self.children:
            if child.key == type and child['name'] == name:
                return child
        return None

    def _FindObject(name,type=''):
        object = None
        if type != '':
            object = AlcScript.current.FindTypedObject(name,type)
        if object == None:
            object = AlcScript.current.FindNamedObject(name,type)
        return object

    def _Initialize():
        print "[AlcScript Parser]"
        if AlcScript.Debug:
            print " Contents of AlcScript:"
            
        blendtext = alcFindBlenderText('AlcScript')
        txt=""
        for line in blendtext.asLines():
            txt=txt + line + "\n"

        text = txt.expandtabs(4) # replace any tabs by 4 spaces, since yaml doesn't accept tabs, and blender tabs are 4 spaces
        AlcScript.current = AlcScript(text,1)
    
    Initialize = staticmethod(_Initialize)
    FindObject = staticmethod(_FindObject)
    

#script = AlcScript("""
#---
#- island:
#    type: object
#    physical:
#        refriction : 5.0
#        elasticity : 0.5
#
#    sound:
#        playtype: loop
#        file: mysound.ogg
#        attenuation: 20.0
#    logic:
#        clickable:
#            condition: InVolumeAndFacing
#            pythonfile: "myaction.py"
#            pyargs:
#                - float: 1.0
#                  
#                - int: 12
#                - int: 0
#            
#- MyCamera1:
#    type: camera
#    transition:
#        speed: 10
#    brain:
#        type: Avatarbrain
#        POA: (1.0,2.0,0.0)
#
#- Lamps:
#    type: object
#    visual:
#        sprite:
#            axis: AxisY
#- Lamps:
#    type: class
#    visual:
#        sprite:
#            axis: AxisY
#---
#""")
#
#obj = script.FindObject("island","object")
#
#if(obj != None):
#    print "Psycho     : ", (obj.Contains('psycho'))
#    print "Logic      : ", (obj.Contains('logic'))
#    if obj.Contains('logic'):
#        print " Clickable : ", (obj.logic.Contains('clickable'))
#        if obj.logic.Contains('clickable'):
#            print "  PyArgs: "
#            print obj.logic.clickable['pyargs']
#else:
#    print "Object 'island' was not in the list"