#
# $Id: alcresmanager.py 862 2007-11-05 23:07:02Z trylon $
#
#    Copyright (C) 2005-2006  Alcugs pyprp Project Team
#    See the file AUTHORS for more info about the team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Please see the file COPYING for the full license.
#    Please see the file DISCLAIMER for more details, before doing nothing.
#

try:
    import Blender
    from Blender import Object
except ImportError:
    pass


import glob, dircache, os
from os.path import *
from binascii import *
import alcconfig
from alc_hsStream import *
from alcm5crypt import *
from alcspecialobjs import *
from alcprpfile import *
from alcurutypes import *
from alc_Classes import *
from alc_AbsClasses import *
from alcGObjects import *
from alcmfsgen import *

HullTypes = {"BOX" : 0, "SPHERE" : 1, "CYLINDER" : 2, "CONE" : 3, "TRIANGLEMESH" : 4, "CONVEXHULL" : 5}


class alcUruPage:
    def __init__(self,parent,name="unnamed",num=0,hide=0,type=0x00):
        self.type="page"
        self.age=parent
        self.resmanager=parent.resmanager
        self.name=name
        self.num=num
        self.hide=hide
        self.type=type
        self.prp=None
        if self.resmanager.filesystem:
            self.ignore=[0x00,]
        else:
            self.ignore=[0x00,0x4C,]
            #self.ignore=[0x00,]
    
    
    def getPath(self,version=None):
        if version==None:
            version=self.age.version
        if version!=6:
            return self.age.base + "/" + self.age.name + "_District_" + self.name + ".prp"
        else:
            return self.age.base + "/" + self.age.name + "_" + self.name + ".prp"
    
    
    def preLoad(self):
        #print " PreLoading %s" %(self.name)
        prpinfo=PrpFileInfo()
        try:
            #f=file(self.getPath(),"rb")
            f=hsStream(self.getPath(),"rb")
        except:
            print "Warning: Not found page %s for age %s" %(self.name,self.age.name)
            return
        prpinfo.read(f)
        f.close()
        #verify seq and num
        if self.age.getSeq()==0:
            raise RuntimeError, "age sequence is not set - but it should be!"
        pid=xPageId()
        pid.setRaw(prpinfo.page_id,prpinfo.page_type,self.age.getSeq())
        self.type=prpinfo.page_type
        pnum=self.num
        if prpinfo.name.name!=self.age.name:
            print "Error: Age name mismatch %s in manifest %s in page" %(self.age.name,prpinfo.name.name)
            print "Page %s in manifest %s in page" %(self.name,prpinfo.page.name)
            print "Debug: Age %s, Page: %s, path: %s, manifest contents:" %(self.age.name,self.name,self.getPath())
            for xpage in self.age.pages:
                print xpage.name, xpage.getPath()
            raise "Error: Age name mismatch %s in manifest %s in page" %(self.age.name,prpinfo.name.name)
        if prpinfo.page.name!=self.name:
            raise "Error: Page name mismatch %s in manifest %s in page" %(self.name,prpinfo.page.name)
        pid2=xPageId()
        pid2.setPage(self.age.getSeq(),pnum,None,prpinfo.page_type)
        if pid2.getRaw()!=pid.getRaw():
            raise "%s Page Id mismatch %08X in manifest, %08X in page" \
                  %(prpinfo.name.name + " " + prpinfo.page.name,pid2.getRaw(),pid.getRaw())

    
    def load(self):
        if self.prp==None:
            print "/\ Loading page %s %i" %(self.name,self.num)
            self.prp=PrpFile(self)
            try:
                f=hsStream(self.getPath(),"rb")
            except IOError:
                self.prp=None
                print "Warning: Not found page %s for age %s" %(self.name,self.age.name)
                return
            self.prp.read(f)
            #self.resmanager.addPrp(self.prp)
            f.close()

    
    def unload(self):
        if self.prp!=None:
            print "\/ Unloading page %s %i" %(self.name,self.num)
            #self.prp.resmanager.delPrp(self.prp)
            del self.prp

    
    def save(self):
        if self.prp==None:
            return
        print "@ Saving page %s %i" %(self.name,self.num)
        #f=file(self.getPath(),"wb")
        f=hsStream(self.getPath(),"wb")
        #self.update_page()
        self.prp.write(f)
        f.close()

    
    def update_page(self):
        thenum=self.num
        if self.type in (0x04,0x00):
            flags=KAgeData
        elif (self.type==0x08 and self.num==-2):
            #self.num=0
            thenum=0
            flags=KAgeSDLHook
        elif (self.type==0x08 and self.num==-1):
            #self.num=0
            thenum=0
            flags=KAgeTextures
        else:
            raise RuntimeError, "Cannot save, Incorrect type of page %i,%i" %(self.type,self.num)
        self.prp.setName(self.age.name)
        self.prp.setPageName(self.name)
        self.prp.setPage(self.age.getSeq(),thenum,flags,0)
        #removed update refs, you may want to turn it on
        ### self.prp.setPage(self.seq,thenum,flags,1) <--

    
    def import_all(self):
        print "-> Importing page %s %i" %(self.name,self.num)
        if not alcconfig.I_agree_that_I_am_NOT_going_to_use_CYAN_material_on_other_games_or_projects_without_written_permission_from_CYAN:
            if self.age.getSeq()<100:
                raise "Import is disabled by default, you must agree that you are not going to use Cyan protected material without their permission"
        if self.prp==None:
            self.load()
            if self.prp==None:
                return

#       base=self.resmanager.getBasePath()
#       agename=self.age.name
#       pagename=self.name
#       et = base + "/" + agename + "/" + pagename
#       try:
#           os.mkdir(et)
#       except OSError:
#           pass
            
        #import
        if self.type in (0x04,0x00):
            #find sceneNode and import all sceneobjects linked to it
            scn = self.prp.getSceneNode()
            if scn==None:
                raise RuntimeError, "This age does not have a Scene Node?"
            
            print "importing from scene node %s" %scn.data.Key.name
            #scn.dump()
            scn.data.import_all()
        elif self.num == -2:
            # Check for an AgeSDLHook
            scenobj = self.prp.find(0x01,"AgeSDLHook",0)
            if scenobj != None:
                # Check for the python file mod
                name = str("VeryVerySpecialPythonFileMod")
                pymod = self.prp.find(0xA2,"VeryVerySpecialPythonFileMod",0)
                if pymod != None:
                    print ">>> AgeSDLHook discovered"
                    try:
                        p = self.age.book.book.getProperty("AgeSDLHook")
                        # Note: Blender does not allow setting of
                        # existing Bool properties; it changes it to an int.  So, we
                        # must remove the existing property and replace it.
                        self.age.book.book.removeProperty(p)
                    except:
                        pass
                    self.age.book.book.addProperty("AgeSDLHook",True,'BOOL')
        #raw
#       for i in self.prp.idx:
#           if i.type in self.ignore:
#              continue
#           
#           a=struct.pack(">H",i.type)
#           b=hexlify(a).upper()
#           cur = et + "/" + b
#           try:
#               os.mkdir(cur)
#           except OSError:
#               pass
#           for o in i.objs:
#               name = str(o.data.Key.name)
#               name = stripIllegalChars(name)
#               #f=file(cur + "/" + name + ".raw","wb")
#               f=hsStream(cur + "/" + name + ".raw","wb")
#               o.write(f)
#               f.close()


    def export_all(self,selection=0):
        print "#########################################"
        print "##"
        print "## => Exporting page %s %i <=" %(self.name,self.num)
        print "##"
        print "#########################################"
        if self.prp==None:
            self.prp=PrpFile(self)
            #set version
            #self.prp.version=self.age.version
            #self.prp.vmin=alcconfig.ver2
            #self.prp.resmanager=self.resmanager
            #self.resmanager.addPrp(self.prp)
            self.update_page()
        out = self.age.base + "/" + self.age.name + "/" + self.name
        SceneNodeRef=UruObjectRef()
        #export
        if self.type in (0x04,0x00):
            #create void sceneNode
            self.prp.createSceneNode()
            self.prp.updateSceneNode()
            SceneNodeRef=self.prp.getSceneNode().data.getRef()
            pass
        elif self.num == -2:
            # Check if the AgeSDLHook property is set to true
            prop = None
            try:
                prop = self.age.book.book.getProperty("AgeSDLHook")
            except:
                # No property - don't do anything
                pass
            if prop != None and bool(prop.getData()) == True:
                # Form a plSceneObject for the hook
                name = str("AgeSDLHook")
                scnobj = self.prp.find(0x01,name,1)
                # Form a plPythonFileMod
                name = str("VeryVerySpecialPythonFileMod")
                pymod = self.prp.find(0xA2,name,1)
                pymod.data.PyFile = Ustr(self.age.name,self.age.version)
                # Give the scene object a ref to the python mod
                scnobj.data.data2.append(pymod.data.getRef())
                print ">>> Added AgeSDLHook scene object and python file mod"

        ##################################################################
        ##                                                              ##
        ## Import of raw objects before export removed at Nov 5 2007    ##
        ##                                                              ##
        ## It's no longer neccessary and just clutters things up        ##
        ##                                                              ##
        ##################################################################

        # First pass: export all soft volumes, since lights may
        # reference them.  Put the other objects in a list.        
        objlist2 = []
        softVolumeParser = alcSoftVolumeParser(self.prp)

        if selection:
            objlist = Blender.Object.GetSelected()
        else:
            objlist = Blender.Object.Get()

        for obj in objlist:
            obj_type=obj.getType()
            if obj_type=="Mesh":
                try:
                    p=obj.getProperty("page_num")
                    if int(p.getData())!=int(self.num):
                        continue
                except (AttributeError, RuntimeError):
                    if self.num!=0:
                        continue
                try:
                    p=obj.getProperty("alctype")
                    alctype=str(p.getData())
                except (AttributeError, RuntimeError):
                    alctype="object"
                
                if alctype == "svconvex":  # Convex Soft Volume
                    #export the object
                    name=str(obj.name)
                    # Create the scene object
                    scnobj=self.prp.find(0x01,name,1)
                    scnobj.data.scene=SceneNodeRef
                    # Create the plSoftVolumeSimple within the scene object
                    softVolume=self.prp.find(0x0088,name,1)
                    softVolume.data.parentref=scnobj.data.getRef()
                    softVolume.data.scenenode=SceneNodeRef
                    softVolume.data.export_object(obj)
                    for svRef in scnobj.data.data1.vector:
                        if svRef.Key.object_type == 0x0088:
                            scnobj.data.data1.remove(svRef)
                    scnobj.data.data1.append(softVolume.data.getRef())
                    softVolumeParser.addSoftVolume(softVolume)

                else: # Not a soft volume - save for second pass
                    objlist2.append(obj)

            else:
                objlist2.append(obj)

        objlist = None

        # Second pass: export all lights and colliders, since the objects may 
        # reference them.  Put the other objects in a list.
        object_list=[]
        clickregion_list=[]
        camregion_list=[]

        for obj in objlist2:
            obj_type=obj.getType()
            if obj_type=="Lamp":
                try:
                    p=obj.getProperty("page_num")
                    if int(p.getData())!=int(self.num):
                        continue
                except (AttributeError, RuntimeError):
                    if self.num!=0:
                        continue
                #export the object
                try:
                    p=obj.getProperty("name")
                    name=str(p.getData())
                except (AttributeError, RuntimeError):
                    name=str(obj.name)

                print "" 
                print "[Lamp %s]" % name


                # --- Obtain scene object ---
                scnobj=self.prp.find(0x01,name,1)
                scnobj.data.scene=SceneNodeRef

                # --- Determine Lamp type and Shadow Type ---
                shadow = None
                if obj.data.type==Blender.Lamp.Types["Spot"]:
                    # plSpotLightInfo
                    lamp=self.prp.find(0x57,name,1)
                    
                    if obj.data.mode & Blender.Lamp.Modes["RayShadow"]:
                        # plPointShadowMaster
                        shadow = self.prp.find(0xD5,name,1)
                elif obj.data.type==Blender.Lamp.Types["Lamp"]:
                    # plOmniLightInfo
                    lamp=self.prp.find(0x56,name,1)
                    if obj.data.mode & Blender.Lamp.Modes["RayShadow"]:
                        # plPointShadowMaster
                        shadow = self.prp.find(0xD5,name,1)
                else:
                    # plDirectionalLightInfo
                    lamp=self.prp.find(0x55,name,1)
                    if obj.data.mode & Blender.Lamp.Modes["RayShadow"]:
                        # plDirectionalShadowMaster
                        shadow = self.prp.find(0xD6,name,1)

                # --- Prepare and Export lamp object ---
                lamp.data.parentref=scnobj.data.getRef()
                lamp.data.scenenode=SceneNodeRef
                lamp.data.softVolumeParser = softVolumeParser
                lamp.data.export_object(obj)
                scnobj.data.data1.append(lamp.data.getRef())

                # --- Prepare and Export Shadow object ---
                if shadow != None:
                    shadow.data.export_obj(obj)
                    scnobj.data.data1.append(shadow.data.getRef())

                # --- Handle the Coordinate interface ---
                coordiface=self.prp.find(0x15,name,1)
                matrix=getMatrix(obj)
                matrix.transpose()
                coordiface.data.matrix.set(matrix)
                coordiface.data.matrix2.set(matrix)
                matrix.invert()
                coordiface.data.matrixinv.set(matrix)
                coordiface.data.matrix2inv.set(matrix)
                coordiface.data.parentref=scnobj.data.getRef()
                scnobj.data.coordinate=coordiface.data.getRef()


            elif obj_type=="Mesh":
                try:
                    p=obj.getProperty("page_num")
                    if int(p.getData())!=int(self.num):
                        continue
                except (AttributeError, RuntimeError):
                    if self.num!=0:
                        continue
                try:
                    p=obj.getProperty("alctype")
                    alctype=str(p.getData())
                except (AttributeError, RuntimeError):
                    alctype="object"
                try:
                    p=obj.getProperty("alcspecial")
                    alcspecial=str(p.getData())
                except (AttributeError, RuntimeError):
                    alcspecial="none"

                if (alctype=="region"):
                    try:
                        p=obj.getProperty("prpregion")
                        prpregion=str(p.getData())
                    except (AttributeError, RuntimeError):
                        prpregion="unknown"
                        
                    if (prpregion == "clickregion"):
                        # click regions should be processed in the first pass
                        # so all click regions are known before we process clickable objects
                        try:
                            p=obj.getProperty("name")
                            name=str(p.getData())
                        except (AttributeError, RuntimeError):
                            name=str(obj.name)

                        print "" 
                        print "[Click Region %s]" % name

                        #find the sceneobject or create it
                        scnobj=self.prp.find(0x01,name,1)
                        scnobj.data.scene=SceneNodeRef
                        
                        # Add it to the click region list
                        clickregion_list.append(scnobj)
                        # Determine the min/max extents of the region
                        bvertexs=obj.getBoundBox()
                        bmaxx=bvertexs[0][0]
                        bmaxy=bvertexs[0][1]
                        bmaxz=bvertexs[0][2]
                        bminx=bvertexs[0][0]
                        bminy=bvertexs[0][1]
                        bminz=bvertexs[0][2]
                        for vertex in bvertexs:
                            if vertex[0]>bmaxx:
                                bmaxx=vertex[0]
                            if vertex[0]<bminx:
                                bminx=vertex[0]
                            if vertex[1]>bmaxy:
                                bmaxy=vertex[1]
                            if vertex[1]<bminy:
                                bminy=vertex[1]
                            if vertex[2]>bmaxz:
                                bmaxz=vertex[2]
                            if vertex[2]<bminz:
                                bminz=vertex[2]
                        #debug print "Click region extents (max X,Y,Z) (min X,Y,Z):"
                        #debug print bmaxx,bmaxy,bmaxz,bminx,bminy,bminz
                        scnobj.vmin=Vertex(bminx,bminy,bminz)
                        scnobj.vmax=Vertex(bmaxx,bmaxy,bmaxz)

                        # col_type will default to 3
                        default_col_type = 3
                        col_type = getStrIntPropertyOrDefault(obj,"col_type",0) #deprecated
                        if col_type == 0:
                            if obj.rbFlags & Object.RBFlags["BOUNDS"]:
                                if obj.rbShapeBoundType == HullTypes["BOX"]:
                                    col_type = 1
                                elif obj.rbShapeBoundType == HullTypes["SPHERE"]:
                                    col_type = 2
                                elif obj.rbShapeBoundType == HullTypes["CYLINDER"]:
                                    col_type = 3
                                elif obj.rbShapeBoundType == HullTypes["CONE"]:
                                    col_type = 3
                                elif obj.rbShapeBoundType == HullTypes["TRIANGLEMESH"]:
                                    col_type = 4
                                elif obj.rbShapeBoundType == HullTypes["CONVEXHULL"]:
                                    col_type = 3
                                else:
                                    col_type = default_col_type
                            else:
                                col_type = default_col_type
                        
                        isdynamic=1 #clickregions are always dynamic (i.e. in need of coordinate interface)
                        
                        # set the simulation interface.
                        self.setSimInterface(scnobj,name,col_type,SceneNodeRef,obj,isdynamic)
    
                        #(re)Set default values, unless the values are defined
                        simi = simi=self.prp.findref(scnobj.data.simulation)
                        phys = simi=self.prp.findref(simi.data.physical)
                        
                        phys.data.mass  = getFloatPropertyOrDefault(obj,"mass",1.0)
                        phys.data.rc    = getFloatPropertyOrDefault(obj,"rc",0.0)
                        phys.data.el    = getFloatPropertyOrDefault(obj,"el",0.0)

                        phys.data.type      = getHexPropertyOrDefault(obj,"col_flags0",2,0x0400)
                        phys.data.flags1    = getHexPropertyOrDefault(obj,"col_flags1",4,0x08000000)
                        phys.data.flags2    = getHexPropertyOrDefault(obj,"col_flags2",4,0x00000000)
                        phys.data.flags3    = getHexPropertyOrDefault(obj,"col_flags3",2,0x0000)
                        phys.data.flags4    = getHexPropertyOrDefault(obj,"col_flags4",4,0x00000004)
                        phys.data.uk3       = getHexPropertyOrDefault(obj,"col_flags5",4,0x00000000)
    
                        # now set the coordinate interface
                        
                        self.setCoordInterface(scnobj,name,isdynamic,obj)
                    
                    elif(prpregion == "camerargn"):
                        # camera regions should also be processed in the first pass
                        # so all click regions are known before we process camera objects
                        try:
                            p=obj.getProperty("name")
                            name=str(p.getData())
                        except (AttributeError, RuntimeError):
                            name=str(obj.name)
                            
                        print "" 
                        print "[Camera Region %s]" % name

                        #find the sceneobject or create it
                        scnobj=self.prp.find(0x01,name,1)
                        scnobj.data.scene=SceneNodeRef
                        
                        # col_type will default to 3
                        default_col_type = 3
                        col_type = getStrIntPropertyOrDefault(obj,"col_type",0) #deprecated way of doing this
                        if col_type == 0:
                            if obj.rbFlags & Object.RBFlags["BOUNDS"]:
                                if obj.rbShapeBoundType == HullTypes["BOX"]:
                                    col_type = 1
                                elif obj.rbShapeBoundType == HullTypes["SPHERE"]:
                                    col_type = 2
                                elif obj.rbShapeBoundType == HullTypes["CYLINDER"]:
                                    col_type = 3
                                elif obj.rbShapeBoundType == HullTypes["CONE"]:
                                    col_type = 3
                                elif obj.rbShapeBoundType == HullTypes["TRIANGLEMESH"]:
                                    col_type = 4
                                elif obj.rbShapeBoundType == HullTypes["CONVEXHULL"]:
                                    col_type = 3
                                else:
                                    col_type = default_col_type
                            else:
                                col_type = default_col_type
                                
                        isdynamic=0 #here is a little problem with camera regions not getting
                                    #transformed, so set it to nondynamic for now
                        
                        # set the simulation interface.
                        self.setSimInterface(scnobj,name,col_type,SceneNodeRef,obj,isdynamic)
                        self.setCoordInterface(scnobj,name,isdynamic,obj)
    
                        #(re)Set default values, unless the values are defined
                        simi = self.prp.findref(scnobj.data.simulation)
                        phys = self.prp.findref(simi.data.physical)
                        
                        phys.data.mass  = getFloatPropertyOrDefault(obj,"mass",1.0)
                        phys.data.rc    = getFloatPropertyOrDefault(obj,"rc",0.0)
                        phys.data.el    = getFloatPropertyOrDefault(obj,"el",0.0)

                        phys.data.type      = getHexPropertyOrDefault(obj,"col_flags0",2,0x0400)
                        phys.data.flags1    = getHexPropertyOrDefault(obj,"col_flags1",4,0x08000000)
                        phys.data.flags2    = getHexPropertyOrDefault(obj,"col_flags2",4,0x00000000)
                        phys.data.flags3    = getHexPropertyOrDefault(obj,"col_flags3",2,0x0000)
                        phys.data.flags4    = getHexPropertyOrDefault(obj,"col_flags4",4,0x00000004)
                        phys.data.uk3       = getHexPropertyOrDefault(obj,"col_flags5",4,0x00000000)
    
                        # read in the camera
                        scnobj.cameralink = getTextPropertyOrDefault(obj,"camera",None)
                        scnobj.blendobj = obj
                        print " Links to Camera '%s'" % (scnobj.cameralink)
    
                        # Add it to the click region list
                        camregion_list.append(scnobj)
                    elif(prpregion == "climbable"):
                        try:
                            p=obj.getProperty("name")
                            name=str(p.getData())
                        except (AttributeError, RuntimeError):
                            name=str(obj.name)

                        print "" 
                        print "[Climbable Region %s]" % name

                        # create ladder regions from this object
                        self.createLadderRegions(obj)
                        # and do not process this region any further
   
                    else: # not a click or camera region
                        object_list.append(obj)
                else: # Not a click region - leave it for later
                    object_list.append(obj)
            else:
                object_list.append(obj)

        objlist2 = None

        #Third pass: export the other objects

        for obj in object_list:
            obj_type=obj.getType()
            if obj_type=="Mesh":
                try:
                    p=obj.getProperty("page")
                    continue
                except (AttributeError, RuntimeError):
                    pass
                try:
                    p=obj.getProperty("book")
                    continue
                except (AttributeError, RuntimeError):
                    pass
                try:
                    p=obj.getProperty("page_num")
                    if int(p.getData())!=int(self.num):
                        continue
                except (AttributeError, RuntimeError):
                    if self.num!=0:
                        continue
                #export the object
                try:
                    p=obj.getProperty("name")
                    name=str(p.getData())
                except (AttributeError, RuntimeError):
                    name=str(obj.name)
                try:
                    p=obj.getProperty("alctype")
                    alctype=str(p.getData())
                except (AttributeError, RuntimeError):
                    alctype="object"
                try:
                    p=obj.getProperty("mass")
                    mass=float(p.getData())
                except (AttributeError, RuntimeError):
                    mass=0.0
                if mass!=0:
                    isdynamic=1
                else:
                    isdynamic=0

                if alctype=="object" or alctype=="sprite": #Object or sprite export
                    #find the sceneobject or create it
                    scnobj=self.prp.find(0x01,name,1)
                    scnobj.data.scene=SceneNodeRef

                  
                    #
                    # Check for sprite
                    #
                    
                                        
                    if(alctype == "sprite"):
                        print "" 
                        print "[Sprite %s]" % name
                        isdynamic = 1 # viewfacemod objects are dynamic too!
                        vfm = self.prp.find(0x40,name + "_VFM",1)
                        vfm.data.export_obj(obj,self.prp)
                        scnobj.data.data2.append(vfm.data.getRef())
                    else:
                        print "" 
                        print "[Visual Object %s]" % name
                    
                    #Determine if we want to generate collision information for this object
                    #
                    col_type = 0
                    col_type = getStrIntPropertyOrDefault(obj,"col_type",0) #deprecated
                    if col_type == 0:
                        if obj.rbFlags & Object.RBFlags["BOUNDS"]:
                            if obj.rbShapeBoundType == HullTypes["BOX"]:
                                print " Using bounds: Bounding Box"
                                col_type = 1
                            elif obj.rbShapeBoundType == HullTypes["SPHERE"]:
                                print " Using bounds: Bounding Sphere"
                                col_type = 2
                            elif obj.rbShapeBoundType == HullTypes["CYLINDER"]:
                                print " Unsupported bounds selected: CYLINDER"
                                print " Defaulting to: Convex Hull"
                                col_type = 3
                            elif obj.rbShapeBoundType == HullTypes["CONE"]:
                                print " Unsupported bounds selected: CONE"
                                print " Defaulting to: Convex Hull"
                                col_type = 3
                            elif obj.rbShapeBoundType == HullTypes["TRIANGLEMESH"]:
                                print " Using bounds: Objects mesh"
                                col_type = 4
                            elif obj.rbShapeBoundType == HullTypes["CONVEXHULL"]:
                                print " Using bounds: Convex Hull"
                                col_type = 3

                    
                    isClimbable = getBoolPropertyOrDefault(obj,"climbable",0)
                    if(isClimbable):
                        # create ladder regions from this object's bounding box
                        self.createLadderRegions(obj)
                        print " -> Object is climbable"
                        print " WARNING: Climbable visual objects must be used with care."
                        print "          Check the following points:"
                        print "          - Object's Y axis points to climbable area"
                        print "          - Object is ladder-sized"
                        print "          If these criteria are not met, unpredictible results might occur"
                        print "          Consider using Climbable regions for more predictability"
                    
                    #
                    # Check For Conditional Objects
                    #
                    try:
                        p=obj.getProperty("sdldefault")
                        sdldefault=bool(p.getData())
                    except (AttributeError, RuntimeError):
                        sdldefault=None
                    if (sdldefault != None):
                        pythname=str("Pyth") + name + str("_Vis")
                        sdlname=name + str("Vis")
                        if sdldefault==True:
                            sdld="1"
                        else:
                            sdld="0"
                        print "=> Conditional Objects (uses and SDL value of 'VAR BOOL	      %s[1]		DEFAULT=%s DEFAULTOPTION=VAULT') <=" % (sdlname, sdld)
                        pythMod = self.prp.find(0x00A2,pythname,1)
                        pythMod.data.PyFile = Ustr("xAgeSDLBoolShowHide",self.age.version)
                        pyArg1 = plPyArg()
                        pyArg1.idx = 0x00000001
                        pyArg1.type = 0x00000004
                        pyArg1.data = str32(sdlname,0)
                        pythMod.data.PyArgs.append(pyArg1)
                        pyArg2 = plPyArg()
                        pyArg2.idx = 0x00000003
                        pyArg2.type = 0x00000001
                        pyArg2.data = 0x00000000
                        pythMod.data.PyArgs.append(pyArg2)
                        scnobj.data.data2.append(pythMod.data.getRef())

                    #
                    # Check for clickable objects
                    #

                    clickfile=getTextPropertyOrDefault(obj,"clickfile",None)
                    if(clickfile != None):
                        print " => Clickable object (uses '%s.py') <=" % clickfile
                        isdynamic = 1 # set the object to dynamic
                        
                        # if col_type is not set (0) set it to 4
                        if (col_type == 0):
                            col_type = 4
                        
                        # Locate a click region that encompasses this object
                        foundClickRegion = None
                        for clickregion in clickregion_list:
                            objX = obj.matrixWorld[3][0]
                            objY = obj.matrixWorld[3][1]
                            objZ = obj.matrixWorld[3][2]
                            #print "Clickable object position: (%f,%f,%f)" %(objX,objY,objZ)
                            if objX >= clickregion.vmin.x and objX <= clickregion.vmax.x \
                              and objY >= clickregion.vmin.y and objY <= clickregion.vmax.y \
                              and objZ >= clickregion.vmin.z and objZ <= clickregion.vmax.z:
                                foundClickRegion = clickregion
                                break
                        if foundClickRegion == None:
                            raise RuntimeError, "(Center of) Clickable object %s is not inside any click region" % name

                        # Create the plPythonFileMod
                        newname = str("cPyth") + name
                        pythMod = self.prp.find(0x00A2,newname,1)
                        pythMod.data.PyFile = Ustr(clickfile,self.age.version)
                        pyArg1 = plPyArg()
                        pyArg1.idx = 0x00000001
                        pyArg1.type = 0x00000007
                        pythMod.data.PyArgs.append(pyArg1)
                        pyArg2 = plPyArg()
                        pyArg2.idx = 0x00000002
                        pyArg2.type = 0x00000004
                        pyArg2.data = str32(name,0)
                        pythMod.data.PyArgs.append(pyArg2)
                        
                        # Create the plPickingDetector
                        newname = str("cClk") + name
                        pickDetect = self.prp.find(0x002B,newname,1)
                        
                        # Create the plObjectInBoxConditionalObject
                        objInBoxCO = self.prp.find(0x0037,newname,1)
                        objInBoxCO.data.bSatisfied = 0x01   # huh, bSatisfied to true?
                        
                        # Create the plActivatorConditionalObject
                        actCO = self.prp.find(0x0032,newname,1)
                        actCO.data.condobjs.append(pickDetect.data.getRef())
                        
                        # Create the plFacingConditionalObject 
                        faceCO = self.prp.find(0x003E,newname,1)
                        faceCO.data.bSatisfied = 0x01   # huh, bSatisfied to true?
                        faceCO.data.fTolerance = -1.0
                        faceCO.data.fDirectional = 0 #false
                        
                        # Create plLogicModifier
                        logicMod = self.prp.find(0x002D,newname,1)
                        logicMod.data.Notify.data.Modifiers.append(pythMod.data.getRef())
                        logicMod.data.fFlags[plLogicModBase.Flags["kLocalElement"]] = 1
                        logicMod.data.fMyCursor = 5 # standard click cursor
                        
                        logicMod.data.Conditionals.append(objInBoxCO.data.getRef())
                        logicMod.data.Conditionals.append(actCO.data.getRef())
                        logicMod.data.Conditionals.append(faceCO.data.getRef())
                        
                        
                        # Create the plInterfaceInfoModifier (same name as object)
                        intInfoMod = self.prp.find(0x00CB,name,1)
                        
                        # Add two references to the same logic modifier
                        intInfoMod.data.objects.append(logicMod.data.getRef())
                        intInfoMod.data.objects.append(logicMod.data.getRef())
                        
                        # Create the plObjectInVolumeDetector 
                        objInVol = self.prp.find(0x007B,newname,1)
                        objInVol.data.logic.append(logicMod.data.getRef())
                        objInVol.data.fType = plCollisionDetector.Type["kTypeAny"]
                        
                        # Hook up the remaining references
                        pyArg1.data = logicMod.data.getRef()
                        pickDetect.data.logic.append(logicMod.data.getRef())
                        clickregion.data.data2.append(objInVol.data.getRef())
                        
                        scnobj.data.data2.append(logicMod.data.getRef())
                        scnobj.data.data2.append(pythMod.data.getRef())
                        scnobj.data.data2.append(pickDetect.data.getRef())
                        scnobj.data.data2.append(intInfoMod.data.getRef())


                    # --- Draw Interface ---
                    drawiref=scnobj.data.draw
                    if drawiref.isNull():
                        drawi=self.prp.find(0x16,name,1)
                        scnobj.data.draw=drawi.data.getRef()
                    else:
                        drawi=self.prp.findref(drawiref)
                    if drawi==None:
                        raise "ERROR: DrawInterface %s - %s not found!" %(str(scnobj.data.Key),str(drawiref))
                        continue
                    # update draw interface
                    drawi.data.parentref=scnobj.data.getRef()
                    # Before exporting the spans, determine whether
                    # the material requires blending; this will affect
                    # the name of the spans
                    RenderLevel = plRenderLevel()
                    ZBias=0
                    
                    # also count the number of texture layers
                    texLayerCount = 0
                    if len(obj.data.materials)>0:
                        materialName = obj.data.materials[0].name
                        pmat=self.prp.find(0x07,materialName,1)
                        
                        # export the material if needed (if it's not done before)
                        ## This will create all texture layers, cubic envmaps and mipmaps, and set the blendings
                        ## If no textures are associated with the material, it will take the (first) uvmap texture of 
                        ## the objects mesh, and use that. (For Backwards compatibility)
                         
                        if(not pmat.isProcessed):
                            pmat.data.export_mat(obj.data.materials[0],obj,self.prp)
                            pmat.isProcessed = 1
                        
                        # get the combined blendflags for this materials layers
                        RenderLevel = pmat.data.RenderLevel()
                        ZBias = pmat.data.ZBias()
                    
                        texLayerCount = pmat.data.TexLayerCount()
                    print "Object's material has %i texture layers" % texLayerCount

                    # --- Drawable Spans
                    
                    # (RenderLevel Level at 0x00000000)
                    SpansLevel = plRenderLevel(plRenderLevel.MajorLevel["kOpaqueMajorLevel"],plRenderLevel.MinorLevel["kDefRendMinorLevel"])
                    if RenderLevel.fLevel == SpansLevel.fLevel:
                        suffix="Spans"
                    else:
                        suffix="BlendSpans"
                    Name_RenderLevel="%08x" % RenderLevel.fLevel
                    Name_ZBias="%x" % ZBias
                    thename=str(self.prp.name) + "_District_" + str(self.prp.page) + "_" + Name_RenderLevel + "_" + Name_ZBias + suffix
                    drawspans=self.prp.find(0x4C,thename,1)
                    drawspans.data.scene=SceneNodeRef
                    drawspans.data.dsiD=RenderLevel.fLevel
                    drawspans.data.dsNumber=ZBias
                    #export the object
                    setnum=drawspans.data.export_object(obj,isdynamic,texLayerCount,self.prp)
                    drawi.data.addSpanSet(setnum,drawspans.data.getRef())

                    # --- Shadow Caster ----
                    
                    if len(obj.data.materials) > 0 :
                        if obj.data.materials[0].mode & Blender.Material.Modes["SHADOWBUF"]:
                            shadowcaster = self.prp.find(0xD4,name,1)
                            shadowcaster.data.export_obj(obj)
                            scnobj.data.data1.append(shadowcaster.data.getRef())


                    # --- Simulation Interface ---
                    self.setSimInterface(scnobj,name,col_type,SceneNodeRef,obj,isdynamic)

                    if(clickfile != None):
                        # set the physical properties correct - but let them get overridden by any set properties

                        simi = simi=self.prp.findref(scnobj.data.simulation)
                        phys = simi=self.prp.findref(simi.data.physical)

                        phys.data.mass  = getFloatPropertyOrDefault(obj,"mass",1.0)
                        phys.data.rc    = getFloatPropertyOrDefault(obj,"rc",0.5)
                        phys.data.el    = getFloatPropertyOrDefault(obj,"el",0.0)

                        phys.data.type      = getHexPropertyOrDefault(obj,"col_flags0",2,0x0400)
                        phys.data.flags1    = getHexPropertyOrDefault(obj,"col_flags1",4,0x00000000)
                        phys.data.flags2    = getHexPropertyOrDefault(obj,"col_flags2",4,0x00020000)
                        phys.data.flags3    = getHexPropertyOrDefault(obj,"col_flags3",2,0x0000)
                        phys.data.flags4    = getHexPropertyOrDefault(obj,"col_flags4",4,0x00000004)
                        phys.data.uk3       = getHexPropertyOrDefault(obj,"col_flags5",4,0x00000002)

                    # -- Coordinate Interface ---
                    self.setCoordInterface(scnobj,name,isdynamic,obj)
                    
                elif alctype=="collider": #Physical export
                    print "" 
                    print "[Collider %s]" % name

                    # allow for climbable colliders
                    isClimbable = getBoolPropertyOrDefault(obj,"climbable",0)
                    if(isClimbable):
                        # create ladder regions from this object's bounding box
                        self.createLadderRegions(obj)
                        print " -> Collider is climbable"
                        print " WARNING: Colliders with climbable addition must be used with care."
                        print "          Please consider using regions of prptype='climbable'"

                    #find the sceneobject or create it
                    scnobj=self.prp.find(0x01,name,1)
                    scnobj.data.scene=SceneNodeRef
                    
                    default_col_type = 3
                    col_type = getStrIntPropertyOrDefault(obj,"col_type",0) #deprecated
                    if col_type == 0:
                        if obj.rbFlags & Object.RBFlags["BOUNDS"]:
                            if obj.rbShapeBoundType == HullTypes["BOX"]:
                                print " Using bounds: Bounding Box"
                                col_type = 1
                            elif obj.rbShapeBoundType == HullTypes["SPHERE"]:
                                print " Using bounds: Bounding Sphere"
                                col_type = 2
                            elif obj.rbShapeBoundType == HullTypes["CYLINDER"]:
                                print " Unsupported bounds selected: CYLINDER"
                                print " Defaulting to: Convex Hull"
                                col_type = 3
                            elif obj.rbShapeBoundType == HullTypes["CONE"]:
                                print " Unsupported bounds selected: CONE"
                                print " Defaulting to: Convex Hull"
                                col_type = 3
                            elif obj.rbShapeBoundType == HullTypes["TRIANGLEMESH"]:
                                print " Using bounds: Objects mesh"
                                col_type = 4
                            elif obj.rbShapeBoundType == HullTypes["CONVEXHULL"]:
                                print " Using bounds: Convex Hull"
                                col_type = 3
                            else:
                                print " Using bounds: Default"
                                col_type = default_col_type
                        else:
                            col_type = default_col_type
                                
                    self.setSimInterface(scnobj,name,col_type,SceneNodeRef,obj)

                elif alctype=="region": #region export
                    print "" 
                    print "[Region Object %s]" % name

                    prpregion=getTextPropertyOrDefault(obj,"prpregion","none")

                    print " Region type: %s"% prpregion
                    
                    #find the sceneobject or create it
                    scnobj=self.prp.find(0x01,name,1)
                    scnobj.data.scene=SceneNodeRef

                    # handle any special col_type settings here:
                    if(prpregion=="swimregion"):
                        default_col_type = 1 # bounding box
                    elif (prpregion == "swimplainsfc") or (prpregion == "swimccursfc") or (prpregion == "swimscursfc"):
                        default_col_type = 4 # objects mesh
                    else:
                        default_col_type = 3 # convex hull

                    col_type = getStrIntPropertyOrDefault(obj,"col_type",0) #deprecated
                    if col_type == 0:
                        if obj.rbFlags & Object.RBFlags["BOUNDS"]:
                            if obj.rbShapeBoundType == HullTypes["BOX"]:
                                col_type = 1
                            elif obj.rbShapeBoundType == HullTypes["SPHERE"]:
                                col_type = 2
                            elif obj.rbShapeBoundType == HullTypes["CYLINDER"]:
                                col_type = 3
                            elif obj.rbShapeBoundType == HullTypes["CONE"]:
                                col_type = 3
                            elif obj.rbShapeBoundType == HullTypes["TRIANGLEMESH"]:
                                col_type = 4
                            elif obj.rbShapeBoundType == HullTypes["CONVEXHULL"]:
                                col_type = 3
                            else:
                                col_type = default_col_type
                        else:
                            col_type = default_col_type

                    isdynamic = 1   #regions are by default dynamic - except when explicitly defined not
                    if(prpregion == "footsteprgn"):

                        # find the "footstepsound" property
                            # default to shallow water,
                            # so they know something is amiss
                        footstepsound = getHexPropertyOrDefault(obj,"footstepsound",2,int(0))
                        footstepsound = getIntPropertyOrDefault(obj,"footstepsound",footstepsound)
                        if(footstepsound > 255): # make sure that the calue doesn't rise out of maximum length
                            footstepsound = 255
                        print " Footstepsound: %i" % footstepsound
                        
                        # find the respmodobject or create it
                        # use the objects name for it
                        ftresp=self.prp.find(0x007C,name,1)
                        # make the feetrespondergroup
                        ftgroup = respmodGroup()
                        ftresp.data.respmodGroup.append(ftgroup)
                        ftaction1 = respmodAction(ftresp.data.getRef())
                        ftgroup.actions.append(ftaction1)
                        ftaction1.flags = 0x00000850
                        ftmessage = PrpMessage(0x038E,self.age.version) # plArmatureEffectStateMsg()
                        ftaction1.message = ftmessage
                        ftaction1.ActionSubType = ftmessage.msgtype
                        ftmessage.data.groundtype = footstepsound
                        
                        # # make the enter objects
                        # make the VolumeSensorConditionalObject
                        vsensor_enter = self.prp.find(0x00A6,name + "_Enter",1)
                        vsensor_enter.data.fType = plVolumeSensorConditionalObject.Type["kTypeEnter"]
                        
                        # make the LogicModifier
                        logicmod_enter = self.prp.find(0x002D,name + "_Enter",1)
                        logicmod_enter.data.fFlags[plLogicModBase.Flags["kMultiTrigger"]] = 1 
                        logicmod_enter.data.Notify.data.Modifiers.append(ftresp.data.getRef())
                        logicmod_enter.data.Conditionals.append(vsensor_enter.data.getRef())
                        
                        objvdetect_enter = self.prp.find(0x007B,name + "_Enter",1)
                        objvdetect_enter.fType = plCollisionDetector.Type["kTypeAny"]
                        objvdetect_enter.data.logic.append(vsensor_enter.data.getRef())
                        objvdetect_enter.data.logic.append(logicmod_enter.data.getRef())
                       
                        # # make the exit objects
                        # make the VolumeSensorConditionalObject
                        vsensor_exit = self.prp.find(0x00A6,name + "_Exit",1)
                        vsensor_exit.data.fType = plVolumeSensorConditionalObject.Type["kTypeExit"]
                        # make the LogicModifier
                        logicmod_exit = self.prp.find(0x002D,name + "_Exit",1)
                        logicmod_exit.data.fFlags[plLogicModBase.Flags["kMultiTrigger"]] = 1 
                        logicmod_exit.data.Notify.data.Modifiers.append(ftresp.data.getRef())
                        logicmod_exit.data.Conditionals.append(vsensor_exit.data.getRef())

                        objvdetect_exit = self.prp.find(0x007B,name + "_Exit",1)
                        objvdetect_enter.fType = plCollisionDetector.Type["kTypeAny"]
                        objvdetect_exit.data.logic.append(vsensor_exit.data.getRef())
                        objvdetect_exit.data.logic.append(logicmod_exit.data.getRef())

                        # # make the interfaceinfomodifier
                        iinfomod = self.prp.find(0x00CB,name,1)
                        iinfomod.data.objects.append(logicmod_enter.data.getRef())
                        iinfomod.data.objects.append(logicmod_exit.data.getRef())

                        # # finally link the
                        # - ResponderModifier 007C 
                        # - LogicModifier 002D "[Regionname]_Enter"       
                        # - LogicModifier 002D "[Regionname]_Exit
                        # - ObjectInVolumeDetector "[Regionname]_Enter"
                        # - ObjectInVolumeDetector "[Regionname]_Exit"
                        # - InterfaceInfoModifier "[Regionname]"
                        # # to the region sceneobject
                        scnobj.data.data2.append(ftresp.data.getRef())
                        scnobj.data.data2.append(logicmod_enter.data.getRef())
                        scnobj.data.data2.append(logicmod_exit.data.getRef())
                        scnobj.data.data2.append(objvdetect_enter.data.getRef())
                        scnobj.data.data2.append(objvdetect_exit.data.getRef())
                        scnobj.data.data2.append(iinfomod.data.getRef())
                    elif(prpregion == "paniclnkrgn"):
                        paniclink = self.prp.find(0x00FC,"PanicLinkReg",1)
                        scnobj.data.data2.append(paniclink.data.getRef())
                    elif(prpregion == "swimrgn"):
                        isdynamic = 0   # Swim Regions are NOT Dynamic
                        swimrgn = self.prp.find(0x012E,name + "_swimrgn",1)
                        swimrgn.parentref = scnobj.data.getRef()
                        swimrgn.parentref2 = scnobj.data.getRef()
                        swimrgn.myself = swimrgn.data.getRef()
                        swimrgn.myself2 = swimrgn.data.getRef()
                        scnobj.data.data2.append(swimrgn.data.getRef())
                    elif(prpregion == "swimplainsfc"):
                        isdynamic = 0   # Swim Surfaces are NOT Dynamic
                        swimsfcrgn = self.prp.find(0x0133,name + "_swimrgnif",1)
                        swimsfcrgn.data.parentref = scnobj.data.getRef()
                        swimsfcrgn.data.export_obj(obj)
                        scnobj.data.data1.append(swimsfcrgn.data.getRef())
                    elif(prpregion == "swimscursfc"): # straight Current // UNFINISHED
                                                 # needs Point object - probably for direction (coord intf only sceneobject)
                        isdynamic = 0   # Swim Surfaces are NOT Dynamic
                        swimsfcrgn = self.prp.find(0x0136,name + "_swimstrcrgn",1)
                        swimsfcrgn.data.parentref = scnobj.data.getRef()
                        swimsfcrgn.data.export_obj(obj,self.prp)
                        scnobj.data.data1.append(swimsfcrgn.data.getRef())
                    elif(prpregion == "swimccursfc"): # circular Current //UNFINISHED
                                                 # needs Point object - probably for direction (coord intf only sceneobject)
                        isdynamic = 0   # Swim Surfaces are NOT Dynamic
                        swimsfcrgn = self.prp.find(0x0134,name + "_swimcircrgn",1)
                        swimsfcrgn.data.parentref = scnobj.data.getRef()
                        swimsfcrgn.data.export_obj(obj,self.prp)
                        scnobj.data.data1.append(swimsfcrgn.data.getRef())
                    elif(prpregion == "climbregion"):
                        #print "Exporting CLIMBREGION"
                        climbRegion = self.prp.find(0x00B2,name,1)
                        if climbRegion != None:
                            #print "On object", name
                            climbRegion.data.parentref = scnobj.data.getRef()
                            climbRegion.data.export_obj(obj,self.prp)
                            scnobj.data.data2.append(climbRegion.data.getRef())
                        else:
                            print "  WARNING: Could not create climbregion"
                        #print "Finished with CLIMBREGION"
                    
                    self.setSimInterface(scnobj,name,col_type,SceneNodeRef,obj,isdynamic)

                    #Set default values, unless the values are defined
                    simi = simi=self.prp.findref(scnobj.data.simulation)
                    phys = simi=self.prp.findref(simi.data.physical)
            
                    if (prpregion == "swimrgn"):
                        phys.data.mass  = getFloatPropertyOrDefault(obj,"mass",0.0)
                        phys.data.rc    = getFloatPropertyOrDefault(obj,"rc",0.0)
                        phys.data.el    = getFloatPropertyOrDefault(obj,"el",0.0)

                        phys.data.type      = getHexPropertyOrDefault(obj,"col_flags0",2,0x0000)
                        phys.data.flags1    = getHexPropertyOrDefault(obj,"col_flags1",4,0x00020000)
                        phys.data.flags2    = getHexPropertyOrDefault(obj,"col_flags2",4,0x00000000)
                        phys.data.flags3    = getHexPropertyOrDefault(obj,"col_flags3",2,0x0000)
                        phys.data.flags4    = getHexPropertyOrDefault(obj,"col_flags4",4,0x00000000)
                        phys.data.uk3       = getHexPropertyOrDefault(obj,"col_flags5",4,0x00000000)
                    elif (prpregion == "swimplainsfc") or (prpregion == "swimccursfc") or (prpregion == "swimscursfc"):
                        phys.data.mass  = getFloatPropertyOrDefault(obj,"mass",0.0)
                        phys.data.rc    = getFloatPropertyOrDefault(obj,"rc",0.0)
                        phys.data.el    = getFloatPropertyOrDefault(obj,"el",0.0)

                        phys.data.type      = getHexPropertyOrDefault(obj,"col_flags0",2,0x0000)
                        phys.data.flags1    = getHexPropertyOrDefault(obj,"col_flags1",4,0x00000000)
                        phys.data.flags2    = getHexPropertyOrDefault(obj,"col_flags2",4,0x00000000)
                        phys.data.flags3    = getHexPropertyOrDefault(obj,"col_flags3",2,0x0000)
                        phys.data.flags4    = getHexPropertyOrDefault(obj,"col_flags4",4,0x00000000)
                        phys.data.uk3       = getHexPropertyOrDefault(obj,"col_flags5",4,0x00000080)
                    elif (prpregion != "unknown") and (prpregion):
                        phys.data.mass  = getFloatPropertyOrDefault(obj,"mass",1.0)
                        phys.data.rc    = getFloatPropertyOrDefault(obj,"rc",0.0)
                        phys.data.el    = getFloatPropertyOrDefault(obj,"el",0.0)

                        phys.data.type      = getHexPropertyOrDefault(obj,"col_flags0",2,0x0400)
                        phys.data.flags1    = getHexPropertyOrDefault(obj,"col_flags1",4,0x08000000)
                        phys.data.flags2    = getHexPropertyOrDefault(obj,"col_flags2",4,0x00000000)
                        phys.data.flags3    = getHexPropertyOrDefault(obj,"col_flags3",2,0x0000)
                        phys.data.flags4    = getHexPropertyOrDefault(obj,"col_flags4",4,0x00000004)
                        phys.data.uk3       = getHexPropertyOrDefault(obj,"col_flags5",4,0x00000000)
                
                                        
                    self.setCoordInterface(scnobj,name,isdynamic,obj)
                        
                ######################
                # Point type support
                ######################
                elif alctype=="point": 
                    print "" 
                    print "[Point Object %s]" % name
                    #find the sceneobject or create it
                    scnobj=self.prp.find(0x01,name,1)
                    scnobj.data.scene=SceneNodeRef
                    # get the coordinates
                    self.setCoordInterface(scnobj,name,1,obj)
#                    coori=self.prp.findref(scnobj.data.coordinate)
#                    if coori==None:
#                        coori=self.prp.find(0x15,name,1)
#                    scnobj.data.coordinate=coori.data.getRef()
#                    coori.data.parentref=scnobj.data.getRef()
#                    coori.data.export_matrices(obj)
            elif obj_type=="Camera":
                try:
                    p=obj.getProperty("page")
                    continue
                except (AttributeError, RuntimeError):
                    pass
                try:
                    p=obj.getProperty("book")
                    continue
                except (AttributeError, RuntimeError):
                    pass
                try:
                    p=obj.getProperty("page_num")
                    if int(p.getData())!=int(self.num):
                        continue
                except (AttributeError, RuntimeError):
                    if self.num!=0:
                        continue
                #export the object
                try:
                    p=obj.getProperty("name")
                    name=str(p.getData())
                except (AttributeError, RuntimeError):
                    name=str(obj.name)
                try:
                    p=obj.getProperty("alctype")
                    alctype=str(p.getData())
                except (AttributeError, RuntimeError):
                    alctype="camera"

                if(alctype == "camera"):
                    print "" 
                    print "[Camera %s]" % name
                    #find the sceneobject or create it
                    scnobj=self.prp.find(0x01,name,1)
                    scnobj.data.scene=SceneNodeRef

                    # --------  Coordinate Interface ----------
                    coori=self.prp.findref(scnobj.data.coordinate)
                    if coori==None:
                        coori=self.prp.find(0x15,name,1)
                    scnobj.data.coordinate=coori.data.getRef()
                    coori.data.parentref=scnobj.data.getRef()
                    coori.data.export_matrices(obj)

                    # --------  Camera Modifier 1 -------------
                    cameramod = self.prp.find(0x009B,name,1)
                    cameramod.data.export_obj(obj,self.prp)
                                     
                    # now link the camera modifier to the object (twice, since that appears to be what cyan does
                    scnobj.data.data2.append(cameramod.data.getRef())
                    scnobj.data.data2.append(cameramod.data.getRef())


                    # --------  Camera Detector Region ---------
                    
                    # Now locate a cameraregion that links to this camera
                    foundCamRegion = 0
                    for camregion in camregion_list:
                        # see what camera regions link to this camera
                        if(camregion.cameralink == name):
                            print "Region %s links to this camera" % str(camregion.data.Key.name)
                            # now create the camera Region Detector
                            camregdetect = self.prp.find(0x006F,name,1)
                            camerakey = scnobj.data.getRef()
                            # find or create the parent object
                            #find the sceneobject or create it
                            camregdetect.data.export_obj(camerakey,camregion.blendobj,self.prp)
                            camregion.data.scene=SceneNodeRef
                            camregion.data.data2.append(camregdetect.data.getRef())
                            foundCamRegion += 1
                    if(foundCamRegion == 0):
                        print "  WARNING: Camera has NO regions associated"
                    else:
                        print " Camera has %d regions associated" % foundCamRegion

            elif obj_type=="Empty":
                try:
                    p=obj.getProperty("page_num")
                    if int(p.getData())!=int(self.num):
                        continue
                except (AttributeError, RuntimeError):
                    if self.num!=0:
                        continue
                #export the object
                name = getTextPropertyOrDefault(obj,"name",str(obj.name))
                alctype = getTextPropertyOrDefault(obj,"alctype","point")

                if alctype=="swpoint": #A spawnPoint
                    print "" 
                    print "[SpawnPoint %s]" % name

                    #find the sceneobject or create it
                    scnobj=self.prp.find(0x01,name,1)
                    scnobj.data.scene=SceneNodeRef
                    #create the spawn modifier
                    found=0
                    for mod in scnobj.data.data2.vector:
                        if mod.Key.object_type==0x3D: #SpawnModifier
                            found=1
                            break
                    if not found:
                        mod=self.prp.find(0x3D,name,1)
                        #print str(mod.data.Key)
                        #print str(mod.data.getRef())
                        scnobj.data.data2.append(mod.data.getRef())
                    # set the coordinate interface
                    coordiface=self.prp.find(0x15,name,1)
                    matrix=getMatrix(obj)
                    matrix.transpose()
                    coordiface.data.matrix.set(matrix)
                    coordiface.data.matrix2.set(matrix)
                    matrix.invert()
                    coordiface.data.matrixinv.set(matrix)
                    coordiface.data.matrix2inv.set(matrix)
                    coordiface.data.parentref=scnobj.data.getRef() 
                    scnobj.data.coordinate=coordiface.data.getRef()
                elif alctype=="point":
                    print "" 
                    print "[Point %s]" % name
                    #find the sceneobject or create it
                    scnobj=self.prp.find(0x01,name,1)
                    scnobj.data.scene=SceneNodeRef
                    # get the coordinates
                    self.setCoordInterface(scnobj,name,1,obj)
                    
#                    coori=self.prp.findref(scnobj.data.coordinate)
#                    if coori==None:
#                        coori=self.prp.find(0x15,name,1)
#                    scnobj.data.coordinate=coori.data.getRef()
#                    coori.data.parentref=scnobj.data.getRef()
#                    coori.data.export_matrices(obj)
                    
                elif alctype=="pymod": #A PythonFileMod
                    print "" 
                    print "[PyMod %s]" % name
                    try:
                        p = obj.getData("PyFile")
                        pyname = str(p.getData())
                    except:
                        raise "You have an unnamed PyMod. Please look into it."
                    found=0
                    for mod in scnobj.data.data2.vector:
                        if mod.Key.object_type==0xA2: #PythonFileMod
                            found=1
                            break
                    if not found:
                        mod=self.prp.find(0xA2,pyname,1)
                        #print str(mod.data.Key)
                        #print str(mod.data.getRef())
                    for i in range(999): #You can have 999 args perfile... Not too shabby
                        try:
                            p = obj.getProperty("PyArg" + i)
                            d = p.getData()
                        except:
                            break #We're done here!
                        mod.data.addArg(i,d)
                ### End export code
        if self.type in (0x04,0x00):
            #update sceneNode
            self.prp.updateSceneNode()
            pass
        #end

    def createLadderRegions(self,obj):
        # creates a set of ladder regions from the bounding box of an object

       
        # get the matrix, rotation and scales
        l2wMtx = getMatrix(obj)
        w2lMtx = Blender.Mathutils.Matrix(l2wMtx).invert()
        
        rotEuler = l2wMtx.rotationPart().toEuler()
        sizeVect = Blender.Mathutils.Vector(obj.size)
    
        # get the objects bounding box
        bbx = obj.getBoundBox()
        
        i = 0
        mworld = []
        mlocal = []
        # order the list like this bitwise, and translate to local coords
        # bit nr   2 1 0
        # index = [x,y,z] -> z/x/y = 1 indicates hightest value, 0 indicates lowest value
        for i in [0,1,3,2,4,5,7,6]: 
            vworld = Blender.Mathutils.Vector(bbx[i][0],bbx[i][1],bbx[i][2],1)
            vlocal = Blender.Mathutils.Vector(vworld) * w2lMtx
    
            mworld.append(vworld)
            mlocal.append(vlocal)
    
    
        # calculate the objects dimensions and center
        deltaLX = (mlocal[7].x - mlocal[0].x)
        deltaLY = (mlocal[7].y - mlocal[0].y)
        deltaLZ = (mlocal[7].z - mlocal[0].z)
        deltaWX = (mlocal[7].x - mlocal[0].x) * sizeVect.x
        deltaWY = (mlocal[7].y - mlocal[0].y) * sizeVect.y
        deltaWZ = (mlocal[7].z - mlocal[0].z) * sizeVect.z
        ctrLX = mlocal[7].x - (deltaLX /2)
        ctrLY = mlocal[7].y - (deltaLY /2)
        ctrLZ = mlocal[7].z - (deltaLZ /2)
        
        # calculate the height of the region:

        # define variables
        rgnheight = 6.0 # region height == avatar height
        rgnwidth = deltaWX # region width is width of the bounding box
        rgndepth = 3.0 # region depth is 3 feet (chosen from educated guess - 2 works probably jst as well...)
        
        ## please do not change the following finely tuned variables, unless you have a very good reason!!!!!
        grabDepth = 0.11 # the amount of space that the avatars hands should penetrate the region, to grab the rungs
        
        gripOffset = 0.50 - grabDepth # put both regions ths amount into the Y+ axis of the bbox, to have it grip the rungs perfectly
        btmYOffset = 0   # bottom region extra offset from local y+ edge of boundbox (offset in Y+ direction)
        topYOffset = -1.00 - grabDepth # top region offset from local y+ edge of boundbox (offset in Y+ direction) 
        btmdepth = rgndepth # Y
        topdepth = rgndepth # optionally we might consider doing rgndepth+deltaWY here. 

    
        # now, we must define a new set of transformation matrices - one for the top region, and one for the bottom region
        
        # now build up the correct Vertex meshes and facelist
        btmMsh = []
        btmMsh.append([0-(rgnwidth/2),0-(btmdepth/2),0])
        btmMsh.append([0-(rgnwidth/2),0-(btmdepth/2),rgnheight])
        btmMsh.append([0-(rgnwidth/2),(btmdepth/2),0])
        btmMsh.append([0-(rgnwidth/2),(btmdepth/2),rgnheight])
        btmMsh.append([(rgnwidth/2),0-(btmdepth/2),0])
        btmMsh.append([(rgnwidth/2),0-(btmdepth/2),rgnheight])
        btmMsh.append([(rgnwidth/2),(btmdepth/2),0])
        btmMsh.append([(rgnwidth/2),(btmdepth/2),rgnheight])
    
        topMsh = []
        topMsh.append([0-(rgnwidth/2),0-(topdepth/2),0])
        topMsh.append([0-(rgnwidth/2),0-(topdepth/2),rgnheight])
        topMsh.append([0-(rgnwidth/2),(topdepth/2),0])
        topMsh.append([0-(rgnwidth/2),(topdepth/2),rgnheight])
        topMsh.append([(rgnwidth/2),0-(topdepth/2),0])
        topMsh.append([(rgnwidth/2),0-(topdepth/2),rgnheight])
        topMsh.append([(rgnwidth/2),(topdepth/2),0])
        topMsh.append([(rgnwidth/2),(topdepth/2),rgnheight])
    
        # now make one face set for both meshes:
        rgnFcs = []
        rgnFcs.append([0,2,6,4])
        rgnFcs.append([5,7,3,1])
        rgnFcs.append([0,1,3,2])
        rgnFcs.append([6,7,5,4])
        rgnFcs.append([4,5,1,0])
        rgnFcs.append([2,3,7,6])
        
        # now calculate where the centers of the top/btm regions should be **locally**, and transform them to world
        # this is a lot easier than taking rotation and stuff into account, and calculate the centers in 
        # world coords directly - nasty thing is that we need to "localize" the top/bottom regions depth (y width)
        
        # first calculate the bottom region's centerpoint, and make it into a vector
        btmVect = Blender.Mathutils.Vector(ctrLX,ctrLY + (gripOffset/sizeVect.y) + (btmYOffset/sizeVect.y) + (deltaLY/2) + ((btmdepth/sizeVect.y)/2), mlocal[0].z,1) * l2wMtx
        # next calculate the top region's centerpoint, and make it into a vector
        topVect = Blender.Mathutils.Vector(ctrLX,ctrLY + (topYOffset/sizeVect.y) + (deltaLY/2) - ((topdepth/sizeVect.y)/2), mlocal[7].z,1) * l2wMtx
        
        # now build up the matrices
        # determine the translation matrices
        btmTrlMtx = Blender.Mathutils.TranslationMatrix(btmVect).resize4x4()
        topTrlMtx = Blender.Mathutils.TranslationMatrix(topVect).resize4x4()
    
        # determine the rotation eulers:
        # default rotation is similar to the objects rotation
        btmEuler = Blender.Mathutils.Euler(rotEuler)
        topEuler = Blender.Mathutils.Euler(rotEuler)
        
        # rotate the top euler by 180 degrees around local Z
        topEuler.rotate(180, 'z')
        
        # make rotation matrices from the eulers
        btmRotMtx = btmEuler.toMatrix().resize4x4()
        topRotMtx = topEuler.toMatrix().resize4x4()
        
        # combine the matrices!
        
        IdentityMtx3d = Blender.Mathutils.Matrix([1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,0])
        btmMtx = btmRotMtx + btmTrlMtx - IdentityMtx3d
        topMtx = topTrlMtx + topRotMtx - IdentityMtx3d
        
        # we now have the following:
        # Bottom region:
        # Matrix btmMtx - VertexList btmMsh - FaceList rgnFcs
        #
        # Top region:
        # Matrix topMtx - VertexList topMsh - FaceList rgnFcs


        # calculate climbHeight and climbOffset
        
        #climbHeight = int(math.ceil((deltaWZ - 5.42)/2))
        climbHeight = int((deltaWZ - 5.42)/2)
        climbOffset = 0
        

        # construct the objects:
        SceneNodeRef=self.prp.getSceneNode().data.getRef()
        # create the bottom region:

        # create the sceneobject
        btmScnobj = self.prp.find(0x01,obj.name + "_bottom",1)
        btmScnobj.data.scene=SceneNodeRef
        # set the coordinates
        btmCoori=self.prp.find(0x15,obj.name + "_bottom",1)
        btmScnobj.data.coordinate=btmCoori.data.getRef()
        btmCoori.data.parentref=btmScnobj.data.getRef()
        btmCoori.data.set_matrices(btmMtx)
        # set the simulation and physical
        btmSimi=self.prp.find(0x1C,obj.name + "_bottom",1)
        btmScnobj.data.simulation=btmSimi.data.getRef()
        btmSimi.data.parentref=btmScnobj.data.getRef()
        btmPhysical=self.prp.find(0x3F,obj.name + "_bottom",1)
        btmSimi.data.physical=btmPhysical.data.getRef()
        btmPhysical.data.scene=SceneNodeRef
        btmPhysical.data.sceneobject=btmScnobj.data.getRef()
        btmPhysical.data.export_synthetic(btmMtx,btmMsh,rgnFcs,4,1)
        # set the ladder specific settings
        btmClimbRegion = self.prp.find(0x00B2,obj.name + "_bottom",1)
        if btmClimbRegion != None:
            btmClimbRegion.data.parentref = btmScnobj.data.getRef()
            # calculate the correct settings
            
            btmClimbRegion.data.export_synthetic(btmMtx,climbHeight,climbOffset,1)
            

            # End calculate the right settings
            btmScnobj.data.data2.append(btmClimbRegion.data.getRef())


        # create the top region:

        # create the sceneobject
        topScnobj = self.prp.find(0x01,obj.name + "_top",1)
        topScnobj.data.scene=SceneNodeRef
        # set the coordinates
        topCoori=self.prp.find(0x15,obj.name + "_top",1)
        topScnobj.data.coordinate=topCoori.data.getRef()
        topCoori.data.parentref=topScnobj.data.getRef()
        topCoori.data.set_matrices(topMtx)
        # set the simulation and physical
        topSimi=self.prp.find(0x1C,obj.name + "_top",1)
        topScnobj.data.simulation=topSimi.data.getRef()
        topSimi.data.parentref=topScnobj.data.getRef()
        topPhysical=self.prp.find(0x3F,obj.name + "_top",1)
        topSimi.data.physical=topPhysical.data.getRef()
        topPhysical.data.scene=SceneNodeRef
        topPhysical.data.sceneobject=topScnobj.data.getRef()
        topPhysical.data.export_synthetic(topMtx,topMsh,rgnFcs,4,1)
        # set the ladder specific settings
        topClimbRegion = self.prp.find(0x00B2,obj.name + "_top",1)
        if topClimbRegion != None:
            topClimbRegion.data.parentref = topScnobj.data.getRef()
            # calculate the correct settings

            topClimbRegion.data.export_synthetic(topMtx,climbHeight,climbOffset,0)

            # End calculate the right settings
            topScnobj.data.data2.append(topClimbRegion.data.getRef())
        

    def setSimInterface(self,scnobj,name,col_type,SceneNodeRef,obj,isdynamic=0):
        if col_type == 0:
            return

        #set the simulation interface
        simiref=scnobj.data.simulation
        simi=self.prp.findref(simiref)
        if simi==None:
            simi=self.prp.find(0x1C,name,1)
        scnobj.data.simulation=simi.data.getRef()
        simi.data.parentref=scnobj.data.getRef()
        physical=self.prp.findref(simi.data.physical)
        if physical==None:
            physical=self.prp.find(0x3F,name,1)
        simi.data.physical=physical.data.getRef()
        physical.data.scene=SceneNodeRef
        physical.data.sceneobject=scnobj.data.getRef()
        physical.data.format=int(col_type)
        physical.data.export_object(obj,isdynamic)

    def setCoordInterface(self,scnobj,name,isdynamic,obj):
        if isdynamic==1:
            #set the coordinate interface
            coori=self.prp.findref(scnobj.data.coordinate)
            if coori==None:
                coori=self.prp.find(0x15,name,1)
            scnobj.data.coordinate=coori.data.getRef()
            coori.data.parentref=scnobj.data.getRef()
            coori.data.export_matrices(obj)
        else:
            scnobj.data.coordinate=UruObjectRef()

class alcUruAge:
    def __init__(self,parent,name="test",basename="./",prefix=0,version=5):
        self.type="age"
        self.resmanager=parent
        self.base=basename
        self.name=name
        self.book=None
        self.version=version
        self.setDefaults(prefix)
    
    
    def setDefaults(self,prefix=None):
        self.options=[]
        self.pages=[]
        if prefix==None:
            prefix=self.getOpt("SequencePrefix")
            if prefix==None:
                prefix=0
        self.mfs=mfs(prefix)
        self.mfs.addFile(self.base + "/" + self.name + ".age")
        self.addOpt("StartDateTime","0000000000")
        self.addOpt("DayLength",24.0)
        self.addOpt("MaxCapacity",150)
        self.addOpt("LingerTime",180)
        self.addOpt("SequencePrefix",prefix)
    
    
    def addOpt(self,name,value):
        if name=="SequencePrefix":
            self.mfs.seq=int(value)
        for me in self.options:
            if me[0].lower()==name.lower():
                me[1]=value
                return
        w=[]
        w.append(name)
        w.append(str(value))
        self.options.append(w)
    
    
    def getOpt(self,name):
        for me in self.options:
            if me[0].lower()==name.lower():
                return me[1]
        return None
    
    
    def findPage(self,name,agename=None):
        for page in self.pages:
            if page.name==name:
                return page
        return None
    
    
    def findPageByNum(self,num):
        for page in self.pages:
            if page.num==num:
                return page
        return None
    
    
    def addPage(self,name,num,hide,type=0x00):
        page=self.findPageByNum(num)
        self.mfs.addFile(self.base + "/" + self.name + "_District_" + name + ".prp")
        if page!=None:
            page.name=name
            page.hide=int(hide)
            page.type=int(type)
            return
        page=alcUruPage(self,name,num,hide,type)
        self.pages.append(page)
    
    
    def getInit(self):
        fpath=self.base + "/" + self.name + ".fni"
        if self.version==6:
            f=M5Crypt()
        else:
            f=Wdys()
        try:
            f.open(fpath,"rb")
            txt=f.read()
            f.close()
        except IOError:
            txt=""
        return txt
    
    
    def setInit(self,txt):
        fpath=self.base + "/" + self.name + ".fni"
        if self.version==6:
            f=M5Crypt()
        else:
            f=Wdys()
        f.open(fpath,"wb")
        f.write(txt)
        f.close()
    
    
    def read(self):
        fpath=self.base + "/" + self.name + ".age"
        f=Wdys()
        try:
            f.open(fpath)
            self.version=5
        except NotWdys:
            try:
                f=M5Crypt()
                f.open(fpath)
                self.version=6
            except NotM5Crypt:
                f=file(fpath,"rb")
                self.version=5
        buf = cStringIO.StringIO()
        buf.write(f.read())
        f.close()
        buf.seek(0)
        IAddedBuiltInPages=0
        for line in buf.readlines():
            w = line.strip().split("=")
            if w[0].lower()!="page":
                self.addOpt(w[0],w[1])
            else:
                if not IAddedBuiltInPages:
                    IAddedBuiltInPages=1
                    self.addBuiltInPages()
                x = w[1].split(",")
                try:
                    flag=int(x[2])
                except IndexError:
                    flag=0
                self.addPage(x[0],int(x[1]),flag)
        #print self.options
        #print self.pages
        buf.close()
        self.preLoadAllPages()
    
    
    def preLoadAllPages(self):
        for page in self.pages:
            page.preLoad()
    
    
    def write(self):
        if self.resmanager.prp_version==6:
            self.f=M5Crypt()
        else:
            self.f=Wdys()
        self.f.open(self.base + "/" + self.name + ".age","wb")
        for opt in self.options:
            self.f.write(opt[0] + "=" + opt[1] + "\r\n")
        for page in self.pages:
            if page.type in (0x04,0x00):
                if page.hide==1:
                    self.f.write("Page=%s,%i,%i\r\n" %(page.name,page.num,page.hide))
                else:
                    self.f.write("Page=%s,%i\r\n" %(page.name,page.num))
        self.f.close()
    
    
    def setSeq(self,val):
        self.addOpt("SequencePrefix",str(val))
    
    
    def getSeq(self):
        return(int(self.getOpt("SequencePrefix")))
    
    
    def addBuiltInPages(self):
        seq=self.getSeq()
        if seq<0:
            return
        else:
            self.addPage("Textures",-1,0,0x08)
            if self.version!=6:
                self.addPage("BuiltIn",-2,0,0x08)

    
    def import_book(self):
        self.book=alcBook(self)
        if self.resmanager.filesystem:
            self.book.save2xml()
        else:
            self.book.find()
            self.book.setAge()
    
    
    def export_book(self,selection):
        self.book=alcBook(self)
        self.book.find(selection)
        self.book.getAge()
    
    
    def import_page(self,name):
        for page in self.pages:
            if page.name==name:
                page.import_all()
                return
        raise "The page %s is not in the %s manifest" %(name,self.name)
    
    
    def find(self,type,name,agename=None):
        for page in self.book.pages:
            if page.prp!=None:
                res=page.prp.find(type,name,0)
                if res!=None:
                    return res
        return None
    
    
    def findref(self,ref):
        if ref.flag==0x01:
            seq=self.getSeq()
            for page in self.pages:
                pid=xPageId()
                pid.setPage(seq,page.num,None,page.type)
                if pid.getRaw()==ref.Key.page_id:
                    if page.prp!=None:
                        return page.prp.findref(ref)
        return None


class alcResManager:
    def __init__(self,datadir="./",version=5,min_ver=12):
        self.type="rmgr"
        self.prp_version=version
        self.prp_min_ver=min_ver
        self.datadir=datadir
        self.ages=[]
        self.filesystem=0
    
    
    def setFilesystem(self):
        self.filesystem=1
    
    
    def getPRPVersion(self):
        return self.prp_version
    
    
    def getBasePath(self):
        return self.datadir
    
    
    def preload(self,datadir=None,age=None):
        if datadir==None:
            datadir=self.datadir
        ages=glob.glob(datadir + "/*.age")
        #print datadir
        #print ages
        for a in ages:
            a=basename(a)[:-4]
            if(age == None or age == a or a[0:6] == 'Global' or a == 'GUI'):
                self.preLoadAge(a,datadir)
            
    
    
    def findAge(self,age,create=0,datadir=None,version=None):
        if version==None:
            version=self.prp_version
        for a in self.ages:
            if a.name==age:
                return a
        if not create:
            return None
        if datadir==None:
            datadir=self.datadir
        a=alcUruAge(self,age,datadir,None,version)
        self.ages.append(a)
        return a
    
    
    def preLoadAge(self,age,datadir=None):
        print "PreLoading %s" %age
        a=self.findAge(age,create=1,datadir=datadir)
        a.read()
    
    
    def import_book(self,agename):
        age=self.findAge(agename)
        if age==None:
            raise "Age %s does not exists!" %(agename)
        age.import_book()
    
    
    def export_book(self,agename,selection):
        age=self.findAge(agename,1)
        age.export_book(selection)

    
    def find(self,type,name,agename=None):
        if agename==None:
            for a in self.ages:
                out=a.find(type,name)
                if out!=None:
                    return out
            return None
        for a in self.ages:
            if a.name==agename:
                return a.find(type,name)
        return None
    
    
    def findref(self,ref):
        if ref.flag==0x01:
            for a in self.ages:
                ret=a.findref(ref)
                if ret!=None:
                    return ret
        return None

    
    def xfindPage(self,pagename,agename=None):
        if agename==None:
            for age in self.ages:
                ret=age.findPage(pagename)
                if ret!=None:
                    return ret
        else:
            age=self.findAge(agename)
            if age==None:
                return None
            return age.findPage(pagename)
        return None

    
    def findPrp(self,name,agename=None):
        for age in self.ages:
            for page in age.pages:
                if page.name==name and page.prp!=None:
                    return page.prp
