# Export stuff

# other needed libraries
import Blender
import prpgui_common
from prpgui_common import *
import alcconfig
import uruprp_export
from os.path import *

# Buttons
EXP_AS_AGE = 501
EXP_AS_PRP = 502
EXP_PPT = 503
EXP_CMP_TEX = 504
EXP_CMP_VRT = 505
EXP_FILEBOX = 506
EXP_BROWSE = 507
EXP_EXPORT = 508

# Variables
export_as_age = Blender.Draw.Create(YES)
export_as_prp = Blender.Draw.Create(NO)
per_page_tex = Blender.Draw.Create(NO)
compress_tex = Blender.Draw.Create(NO)
compress_verts = Blender.Draw.Create(NO)
export_name = Blender.Draw.Create("")

# the file browser callback
def file_callback(filename):
	global debug, verbose_debug
	global export_name

	if verbose_debug:
		print "Entering 'File Browser' callback."
	
	if export_as_age:
		ext = ".age"
	else:
		ext = ".prp"
	if filename.find(ext,-4) == -1:
		Blender.Draw.PupMenu("Unsuported file %s, expecting an %s file" %(filename,ext))
	export_name.val = filename
	if debug:
		print "Filename set to: %s" % filename

# The export function
	# This function sets up a few things, then calls the export function
	# from the old PyPRP system.
def export_age(filename):
	global debug, verbose_debug
	global export_as_age, export_as_prp, per_page_tex, compress_tex
	global compress_verts, export_name
	global YES, NO

	if verbose_debug:
		print "Entering 'Export Age' function."
	
	alcconfig.texture_compression = compress_tex.val
	alcconfig.vertex_compression = compress_verts.val
	alcconfig.export_textures_to_page_prp = per_page_tex.val
	basepath = dirname(filename)
	if export_as_age.val:
		agename = basename(filename[:-4])
		uruprp_export.export_age(agename,basepath,0,0)
	else:
		pagea = basename(filename[:-4])
		page = pagea.split('_')
		agename = page[0]
		pagename = pagea[len(page[0]) + 1 + len(page[1]) + 1:]
		uruprp_export.export_age(agename,basepath,0,0,pagename)
		

# The button event handler
def handler(evt):
	global debug, verbose_debug
	global export_as_age, export_as_prp, per_page_tex, compress_tex
	global compress_verts, export_name
	global YES, NO
	global EXP_AS_AGE, EXP_AS_PRP, EXP_PPT, EXP_CMP_TEX, EXP_CMP_VRT
	global EXP_FILEBOX, EXP_BROWSE, EXP_EXPORT

	if verbose_debug:
		print "Entering 'Export' Button handler."
	
	if evt == EXP_AS_AGE:
		if debug:
			print "Export type set to: Age"
		export_as_age.val = YES
		export_as_prp.val = NO
	elif evt == EXP_AS_PRP:
		if debug:
			print "Export type set to: PRP"
		export_as_age.val = NO
		export_as_prp.val = YES
	elif evt == EXP_PPT:
		if debug:
			if per_page_tex:
				print "Per-page texture storage enabled"
			else:
				print "Per-page texture storage disabled"
	elif evt == EXP_CMP_TEX:
		if debug:
			if compress_tex:
				print "Texture compression enabled"
			else:
				print "Texture compression disabled"
	elif evt == EXP_CMP_VRT:
		if debug:
			if compress_verts:
				print "Vertex compression Enabled"
			else:
				print "Vertex compression disabled"
	elif evt == EXP_FILEBOX:
		if debug:
			print "Filename set to: %s" % export_name.val
	elif evt == EXP_BROWSE:
		if debug:
			print "Bringing up file browser"
		Blender.Window.FileSelector(file_callback, 'Select File', export_name.val)
	elif evt == EXP_EXPORT:
		if debug:
			print "Exporting Age"
		export_age(export_name.val)
	else:
		print "Unknown button: %d" % evt
		Blender.Draw.Exit()

def draw():
	global debug, verbose_debug
	global export_as_age, export_as_prp, per_page_tex, compress_tex
	global compress_verts, export_name
	global YES, NO
	global EXP_AS_AGE, EXP_AS_PRP, EXP_PPT, EXP_CMP_TEX, EXP_CMP_VRT
	global EXP_FILEBOX, EXP_BROWSE, EXP_EXPORT

	if verbose_debug:
		print "Entering 'Export' Draw function."
	
	Blender.BGL.glRasterPos2i(135, 380)
	Blender.Draw.Text("Export As:")
	export_as_age = Blender.Draw.Toggle("Age", EXP_AS_AGE, 195, 375, 40, 20, export_as_age.val, "Export a full age")
	export_as_prp = Blender.Draw.Toggle("PRP", EXP_AS_PRP, 240, 375, 40, 20, export_as_prp.val, "Export a single PRP")
	per_page_tex = Blender.Draw.Toggle("Per-Page Textures", EXP_PPT, 180, 330, 120, 20, per_page_tex.val, "Export textures on the pages they're used.")
#	compress_verts = Blender.Draw.Toggle("Vertex Compression", EXP_CMP_VRT, 165, 300, 135, 20, compress_verts.val, "Compress vertex data on export.")
	# Vertex Compression broken. Turned off until further notice.
	compress_tex = Blender.Draw.Toggle("Texture Compresion", EXP_CMP_TEX, 160, 270, 140, 20, compress_tex.val, "Compress texture data on export.")
	Blender.BGL.glRasterPos2i(20, 85)
	Blender.Draw.Text("Filename:")
	export_name = Blender.Draw.String("", EXP_FILEBOX, 20, 60, 215, 20, export_name.val, 100)
	Blender.Draw.PushButton("Browse", EXP_BROWSE, 240, 60, 60, 20, "Browse for the file to export to.")
	Blender.Draw.PushButton("Export", EXP_EXPORT, 120, 30, 80, 20, "Export the Age")

	
