# Main properties stuff. Anything not specific to a single object type

import Blender
import prpgui_common
from prpgui_common import *

# Buttons (global)
PRP_NAME = 301
PRP_PAGE = 302
PRP_COL_TYP = 303
PRP_COPY = 304
PRP_PASTE = 305

# Buttons (Regions)
PRP_RGN_TYP = 306
PRP_STRT_DIS = 307
PRP_STATIC = 308
PRP_FOOT_SND = 309
PRP_LDDR_RGN = 310
PRP_BTTM_FLG = 311
PRP_CNVRT_OBJ = 312

# Buttons (Objects)
PRP_SAV_STAT = 313
PRP_BLK_CAM = 314
PRP_NO_FRIC = 315
PRP_CLICK = 316
PRP_COLLIDE = 317
PRP_VISIBLE = 318
PRP_WTH_ALL = 319
PRP_WTH_KCK = 320
PRP_MASS = 321
PRP_RC = 322
PRP_EL = 323
PRP_SCRIPT = 324
PRP_CNVRT_RGN = 325

# Defines (Regions)
RGN_TYP_FOOTSTEP = 0
RGN_TYP_PANIC = 1
RGN_TYP_SWIM = 2
RGN_TYP_SWM_SFC = 3
RGN_TYP_CURRENT = 4
RGN_TYP_CCURRENT = 5
RGN_TYP_CLICK = 6
RGN_TYP_CAM = 7
RGN_TYP_CLIMB = 8
REGION_TYPES = ['footsteprgn','paniclnkrgn','swimrgn','swimplainsfc','swimscursfc','swimccursfc','clickregion','camerargn','climbregion']

# Defines (Properties)
CF0_SV_STAT = 0x0100
CF0_NO_SAVE = 0x0200
CF2_COLLIDE_ALL = 0x02000000
CF2_COLLIDE_KCK = 0x01000000
CF4_PINNED = 0x00000004
CF4_CAM_AVOID = 0x00000080
CF4_SUPPRESS = 0x00000800
CF5_CAM_AVOID = 0x00000004
CF5_NO_FRIC = 0x00000040

# Variables (Global)
prp_name = Blender.Draw.Create("")
prp_page = Blender.Draw.Create(0)
col_type = Blender.Draw.Create(1)

# Variables (Regions)
rgn_typ = Blender.Draw.Create(RGN_TYP_FOOTSTEP)
start_disabled = Blender.Draw.Create(NO)
static = Blender.Draw.Create(YES)
stp_snd = Blender.Draw.Create(19)

# Variables (Objects)
save_state = Blender.Draw.Create(NO)
block_cam = Blender.Draw.Create(YES)
no_fric = Blender.Draw.Create(NO)
clickable = Blender.Draw.Create(NO)
collider = Blender.Draw.Create(YES)
visible = Blender.Draw.Create(YES)
col_wth_all = Blender.Draw.Create(YES)
col_wth_kck = Blender.Draw.Create(NO)
mass = Blender.Draw.Create(0.00)
el = Blender.Draw.Create(0.00)
rc = Blender.Draw.Create(0.00)
script = Blender.Draw.Create("")

# Variables (Properties)
col_flags0 = 0x0100
col_flags1 = 0x00000000
col_flags2 = 0x02000000
col_flags3 = 0x0000
col_flags4 = 0x00000080
col_flags5 = 0x00000004

def region_handler(evt):
	global debug, verbose_debug, curobj
	global rgn_typ, start_disabled, static, stp_snd
	global col_flags4
	global YES, NO, REGION_TYPES, SOUND_TYPES
	global PRP_RGN_TYP, PRP_STRT_DIS, PRP_STATIC, PRP_FOOT_SND
	global CF4_SUPPRESS

	if verbose_debug:
		print "Entering 'Region Properties' handler."

	if evt == PRP_RGN_TYP:
		if debug:
			print "Region type changed to: %s" % REGION_TYPES[rgn_typ.val]
		prop = curobj.getProperty('prpregion')
		prop.setData(REGION_TYPES[rgn_typ.val])
	elif evt == PRP_STRT_DIS:
		prop = curobj.getProperty('col_flags4')
		if start_disabled.val:
			if debug:
				print "'Start Disabled' enabled."
			col_flags4 = col_flags4 | CF4_SUPPRESS
		else:
			if debug:
				print "'Start Disabled' disabled."
			col_flags4 = col_flags4 & (~CF4_SUPPRESS)
		prop.setData("%08x" % col_flags4)
	elif evt == PRP_STATIC:
		prop = curobj.getProperty('mass')
		if static.val:
			if debug:
				print "'Static' enabled."
			prop.setData(0.00)
		else:
			if debug:
				print "'Static' disabled."
			prop.setData(1.00)
	elif evt == PRP_FOOT_SND:
		if debug:
			print "Footstep sound set to: %d" % stp_snd.val
		prop = curobj.getProperty('footstepsnd')
		prop.setData(stp_snd.val)

def object_handler(evt):
	global debug, verbose_debug, curobj
	global YES, NO
	global save_state, block_cam, no_fric, clickable, collider, visible
	global col_wth_all, col_wth_kck, mass, el, rc, script
	global PRP_SAV_STAT, PRP_BLK_CAM, PRP_NO_FRIC, PRP_CLICK, PRP_COLLIDE
	global PRP_VISIBLE, PRP_WTH_ALL, PRP_WTH_KCK, PRP_MASS, PRP_RC, PRP_EL
	global PRP_SCRIPT, PRP_CNVRT_RGN

	if verbose_debug:
		print "Entering 'Object Properties' handler."
	
	if evt == PRP_SAV_STAT:
		prop = curobj.getProperty('col_flags0')
		if save_state.val:
			if debug:
				print "'Save Object State' enabled."
			col_flags0 = col_flags0 | CF0_SV_STAT
			col_flags0 = col_flags0 & (~CF0_NO_SAVE)
		else:
			if debug:
				print "'Save Object State' disabled."
			col_flags0 = col_flags0 | CF0_NO_SAVE
			col_flags0 = col_flags0 & (~CF0_SV_STAT)
		prop.setData("%04x" % col_flags0)
	elif evt == PRP_BLK_CAM:
		prop = curobj.getProperty('col_flags5')
		prop2 = curobj.getProperty('col_flags4')
		if block_cam.val:
			if debug:
				print "'Camera Blocking' enabled."
			col_flags5 = col_flags5 | CF5_CAM_AVOID
			col_flags4 = col_flags4 | CF4_CAM_AVOID
		else:
			if debug:
				print "'Camera Blocking' disabled."
			col_flags5 = col_flags5 & (~CF5_CAM_AVOID)
			col_flags4 = col_flags4 & (~CF4_CAM_AVOID)
		prop.setData("%08x" % col_flags5)
		prop2.setData("%08x" % col_flags4)
	elif evt == PRP_NO_FRIC:
		prop = curobj.getProperty('col_flags5')
		if no_fric.val:
			if debug:
				print "'Friction With Avatar' disabled."
			col_flags5 = col_flags5 | CF5_NO_FRIC
		else:
			if debug:
				print "'Friction With Avatar' enabled."
			col_flags5 = col_flags5 & (~CF5_NO_FRIC)
		prop.setData("%08x" % col_flags5)
	elif evt == PRP_CLICK:
		if clickable.val:
			if debug:
				print "'Clickable' enabled."
		else:
			if debug:
				print "'Clickable' disabled."

def handler(evt):
	global debug, verbose_debug, curobj
	global prp_name, prp_page, col_type
	global YES, NO, COL_TYPES
	global PRP_NAME, PRP_PAGE, PRP_COL_TYP, PRP_COPY, PRP_PASTE

	if verbose_debug:
		print "Entering 'Properties' handler."
	
	if evt == PRP_NAME:
		if debug:
			print "Object/Region name changed to: %s" % prp_name.val
		prop = curobj.getProperty('name')
		prop.setData(prp_name.val)
	elif evt == PRP_COL_TYP:
		if debug:
			print "Collision type changed to: %d" % col_type.val
		prop = curobj.getProperty('col_type')
		prop.setData("%d" % col_type.val)
	
	prop = curobj.getProperty('alctype')
	type = prop.getData()
	if type == 'region':
		region_handler(evt)
	elif type == 'object':
		object_handler(evt)

def draw_region():
	global debug, verbose_debug
	global rgn_typ, start_disabled, static, stp_snd
	global YES, NO
	global PRP_RGN_TYP, PRP_STRT_DIS, PRP_STATIC, PRP_FOOT_SND
	global RGN_TYP_FOOTSTEP, RGN_TYP_PANIC, RGN_TYP_CLICK

	if verbose_debug:
		print "Entering 'Region Properties' Draw function."
	if rgn_typ.val != RGN_TYP_FOOTSTEP and rgn_typ.val != RGN_TYP_PANIC and rgn_typ.val != RGN_TYP_CLICK:
		Blender.BGL.glRasterPos2i(20, 300)
		Blender.Draw.Text("This region type isn't supported yet!")
		Blender.BGL.glRasterPos2i(20, 270)
		Blender.Draw.Text("You'll have to set its properties manually.")
		return

	regions = "Region Type %t|Footstep %x0|Panic Link %x1|Click %x6|"
	rgn_typ = Blender.Draw.Menu(regions, PRP_RGN_TYP, 20, 330, 115, 20, rgn_typ.val, "The type of Region.")
	start_disabled = Blender.Draw.Toggle("Start Disabled", PRP_STRT_DIS, 195, 300, 105, 20, start_disabled.val, "Region won't work until turned on in a script.")
	static = Blender.Draw.Toggle("Static", PRP_STATIC, 245, 220, 55, 20, static.val, "Region cannot be moved, even by a script. Slight Performance boost possible, but unconfirmed")
	if rgn_typ.val == RGN_TYP_FOOTSTEP:
		sounds = "Sound %t|Dirt %x0|Puddle %x1|Water %x2|Tile %x3|Metal %x4|Wooden Bridge %x5|Rope Ladder %x6|Grass %x7|Brush %x8|Hard Wood %x9|Rug %x10|Stone %x11|Mud %x12|Metal Ladder %x13|Wooden Ladder %x14|Deep Water %x15|Maintainer Glass %x16|Maintainer Stone %x17|Swimming %x18|None %x19"
		stp_snd = Blender.Draw.Menu(sounds, PRP_FOOT_SND, 20, 210, 135, 20, stp_snd.val, "The sound the footsteps make.")

def draw():
	global debug, verbose_debug, curobj
	global prp_name, prp_page, col_type
	global YES, NO
	global PRP_NAME, PRP_PAGE, PRP_COL_TYP, PRP_COPY, PRP_PASTE

	if verbose_debug:
		print "Entering 'Properties' Draw function."
	
	if curobj == None:
		Blender.BGL.glRasterPos2i(20, 400)
		Blender.Draw.Text("Must select a mesh object!")
		return
	
	prop = curobj.getProperty("alctype")
	type = prop.getData()
	
	Blender.BGL.glRasterPos2i(20, 385)
	Blender.Draw.Text("Name:")
	if prp_name.val == None:	# Don't know where this comes from, but it's a pain in the arse. Fix is pretty easy, though...
		prop = curobj.getProperty("name")	# Just grab the property again :)
		prp_name.val = prop.getData()
	prp_name = Blender.Draw.String("", PRP_NAME, 65, 380, 95, 20, prp_name.val, 20, "The name of the object.")
	collision = "Collision Type %t|Bounding Box %x1|Bounding Sphere %x2|Convex Hull %x3|Triangles %x4|Terrain %x5"
	col_type = Blender.Draw.Menu(collision, PRP_COL_TYP, 175, 330, 125, 20, col_type.val, "Method use for collision detection.")
	
	if type == 'region':
		draw_region()
	else:
		Blender.BGL.glRasterPos2i(20, 300)
		Blender.Draw.Text("Not yet implemented.")
		return

def obj_setup():
	global debug, verbose_debug, curobj
	global prp_name, prp_page, col_type
	global save_state, block_cam, no_fric, clickable, collider, visible
	global col_wth_all, col_wth_kck, mass, el, rc, script
	global rgn_typ, start_disabled, static, stp_snd
	global col_flags0, col_flags1, col_flags2, col_flags3, col_flags4, col_flags5
	global YES, NO
	global RGN_TYP_FOOTSTEP, RGN_TYP_PANIC, RGN_TYP_SWIM, RGN_TYP_SWM_SFC
	global RGN_TYP_CURRENT, RGN_TYP_CCURRENT, RGN_TYP_LADDER, RGN_TYP_CLIMB
	global RGN_TYP_CAM, RGN_TYP_CLICK
	global CF0_SV_STAT, CF0_NO_SAVE
	global CF2_COLLIDE_ALL, CF2_COLLIDE_KCK
	global CF4_PINNED, CF4_CAM_AVOID, CF4_SUPPRESS
	global CF5_CAM_AVOID, CF5_NO_FRIC
	# This function needs to do a lot of things.
		# 1) Figure out what we're dealing with. Object,Region,Collider,Sprite,etc.
		# 2) Figure out how the currently set properties map to the button layout
		# 3) Warn about un-supported options
		# 4) Decide whether or not to deal with this particular object.
	if verbose_debug:
		print "Entering 'Object Setup' function."
	
	Scene = Blender.Scene.GetCurrent()
	curobj = Scene.getActiveObject()
	type = curobj.getType()
	
	if type == 'Camera':	# Cameras get a nice little note saying 'Not yet supported.'
		return
	elif type != 'Mesh':
		curobj = None
		return
	
	object_name = curobj.getName()
	
	if object_name.startswith('xPage') or object_name == 'xBook':	# Skip the page and book stuff.
		curobj = None
		return
	
	# Now for the fun bit. Every single property needs to be checked.
		# If a property exists that we use, the proper information is passed on to the GUI
		# If a property doesn't exist that we use, reasonable defaults are set
		# If a property exists that we don't use, it's ignored
		# If a property exists that we use, but has an invalid value, it's adjusted.
			# NOTE: bit-fields are special. Any un-used bits will be ignored.
	is_region = NO
	if debug:
		print "Checking 'name' property."
	try:
		prop = curobj.getProperty('name')
	except:
		curobj.addProperty('name',object_name,'STRING')
		prp_name.val = object_name
	else:
		if prop.getType != 'STRING':
			prop.setData(object_name)
			prp_name.val = object_name
		else:
			prp_name.val = prop.getData()
	if debug:
		print "Checking 'col_type' property."
	try:
		prop = curobj.getProperty('col_type')
	except:
		curobj.addProperty('col_type','1','STRING')
		col_type.val = 1
	else:
		col_type.val = int(prop.getData())
		if col_type.val > 5:
			col_type.val = 1
			prop.setData('1')
	if debug:
		print "Checking 'alctype' property."
	try:
		prop = curobj.getProperty('alctype')
	except:
		curobj.addProperty('alctype','object','STRING')
	else:
		if prop.getType() != 'STRING':
			prop.setData('object')
		else:
			type = prop.getData()
			if type == 'sprite' or type == 'swpoint' or type == 'point':	# Non-implemented types.
				return
			elif type == 'region':
				is_region = YES
			elif type == 'collider':
				visible.val = NO		# could check this later, but no sense re-grabbing the property
			elif type == 'object':
				visible.val = YES
			else:
				prop.setData('object')
				visible.val = YES
	if is_region:
		if debug:
			print "Checking 'prpregion' property."
		try:
			prop = curobj.getProperty('prpregion')
		except:
			curobj.addProperty('prpregion','footsteprgn','STRING')
			rgn_typ.val = RGN_TYP_FOOTSTEP
			type = 'footsteprgn'
		else:
			if prop.getType() != 'STRING':
				prop.setData('footsteprgn')
				type = 'footsteprgn'
				rgn_typ.val = RGN_TYP_FOOTSTEP
			else:
				type = prop.getData()
				if type == 'footsteprgn':
					rgn_typ.val = RGN_TYP_FOOTSTEP
				elif type == 'paniclnkrgn':
					rgn_typ.val = RGN_TYP_PANIC
				elif type == 'swimrgn':
					rgn_typ.val = RGN_TYP_SWIM
					return
				elif type == 'swimplainsfc':
					rgn_typ.val = RGN_TYP_SWM_SFC
					return
				elif type == 'swimscursfc':
					rgn_typ.val = RGN_TYP_CURRENT
					return
				elif type == 'swimccursfc':
					rgn_typ.val = RGN_TYP_CCURRENT
					return
				elif type == 'climbregion':
					rgn_typ.val = RGN_TYP_CLIMB
					return
				elif type == 'camerargn':
					rgn_typ.val = RGN_TYP_CAM
					return
				elif type == 'clickregion':
					rgn_typ.val = RGN_TYP_CLICK
				else:	# any regions that are unknown get set to footstep... probably a bad idea, but meh.
					prop.setData('footsteprgn')
					type = 'footsteprgn'
					rgn_typ.val = RGN_TYP_FOOTSTEP
		if type == 'footsteprgn':			# only gets checked if we're already a region
			if debug:
				print "Checking 'footstepsnd' property."
			try:
				prop = curobj.getProperty('footstepsnd')
			except:
				curobj.addProperty('footstepsnd',19,'INT')
				stp_snd.val = 19
			else:
				if prop.getType() != 'INT':
					prop.setData(19)
					stp_snd.val = 19
				else:
					stp_snd.val = prop.getData()
					if stp_snd.val > 19:
						stp_snd.val = 19
						prop.setData(19)
		if debug:
			print "Checking 'mass' property."
		try:
			prop = curobj.getProperty('mass')
		except:
			curobj.addProperty('mass',0.00,'FLOAT')
			static.val = YES
		else:
			if prop.getType != 'FLOAT':
				prop.setData(0.00)
				static.val = YES
			else:
				if prop.getData == 0.00:
					static.val = YES
				else:
					static.val = NO
		if debug:
			print "Checking 'col_flags0' property."
		try:
			prop = curobj.getProperty('col_flags0')
		except:
			curobj.addProperty('col_flags0','0200','STRING')
		else:
			prop.setData('0200')
		col_flags0 = 0x0200
		if debug:
			print "Checking 'col_flags1' property."
		try:
			prop = curobj.getProperty('col_flags1')
		except:
			curobj.addProperty('col_flags1','00000000','STRING')
		else:
			prop.setData('00000000')
		col_flags1 = 0x00000000
		if debug:
			print "Checking 'col_flags2' property."
		try:
			prop = curobj.getProperty('col_flags2')
		except:
			curobj.addProperty('col_flags2','02000000','STRING')
		else:
			prop.setData('02000000')
		col_flags2 = 0x02000000
		if debug:
			print "Checking 'col_flags3' property."
		try:
			prop = curobj.getProperty('col_flags3')
		except:
			curobj.addProperty('col_flags3','0000','STRING')
		else:
			prop.setData('0000')
		col_flags3 = 0x0000
		if debug:
			print "Checking 'col_flags4' property."
		try:
			prop = curobj.getProperty('col_flags4')
		except:
			curobj.addProperty('col_flags4','00000004','STRING')
			start_disabled.val = NO
		else:
			if prop.getType() != 'STRING':
				prop.setData('00000004')
				start_disabled.val = NO
			else:
				col_flags4 = int(prop.getData(),16)
				if not col_flags4 & CF4_PINNED:
					col_flags4 = col_flags4 | CF4_PINNED
				if col_flags4 & CF4_SUPPRESS:
					start_disabled.val = YES
				prop.setData('%08x' % col_flags4)
		if debug:
			print "Checking 'col_flags5' property."
		try:
			prop = curobj.getProperty('col_flags5')
		except:
			curobj.addProperty('col_flags5','00000000','STRING')
		else:
			prop.setData('00000000')
	else:	# objects/colliders. Will add a section for other types as they are implemented
		if debug:
			print "Checking 'mass' property."
		try:
			prop = curobj.getProperty('mass')
		except:
			curobj.addProperty('mass',0.00,'FLOAT')
			mass.val = 0.00
		else:
			if prop.getType() != 'FLOAT':
				prop.setData(0.00)
				mass.val = 0.00
			else:
				mass.val = prop.getData()
		if debug:
			print "Checking 'rc' property."
		try:
			prop = curobj.getProperty('rc')
		except:
			curobj.addProperty('rc',0.00,'FLOAT')
			rc.val = 0.00
		else:
			if prop.getType() != 'FLOAT':
				prop.setData(0.00)
				rc.val = 0.00
			else:
				rc.val = prop.getData()
		if debug:
			print "Checking 'el' property."
		try:
			prop = curobj.getProperty('el')
		except:
			curobj.addProperty('el',0.00,'FLOAT')
			el.val = 0.00
		else:
			if prop.getType() != 'FLOAT':
				prop.setData(0.00)
				el.val = 0.00
			else:
				el.val = prop.getData()
		if debug:
			print "Checking 'col_flags0' property."
		try:
			prop = curobj.getProperty('col_flags0')
		except:
			curobj.addProperty('col_flags0','0200','STRING')
			save_state.val = NO
			col_flags0 = 0x0200
		else:
			if prop.getType() != 'STRING':
				prop.setData('0200')
				save_state.val = NO
				col_flgas0 = 0x0200
			else:
				col_flags0 = int(prop.getData(),16)
				if col_flags0 & CF0_SV_STAT:
					save_state.val = YES
					col_flags0 = col_flags0 & (~CF0_NO_SAVE)
				else:
					save_state.val = NO
					col_flags0 = col_flags0 & (~CF0_SV_STAT)
					col_flags0 = col_flags0 | CF0_NO_SAVE
		if debug:
			print "Checking 'col_flags1' property."
		try:
			prop = curobj.getProperty('col_flags1')
		except:
			curobj.addProperty('col_flags1','00000000','STRING')
		else:
			prop.setData('00000000')
		col_flags1 = 0x00000000
		if debug:
			print "Checking 'col_flags2' property."
		try:
			prop = curobj.getProperty('col_flags2')
		except:
			curobj.addProperty('col_flags2','02000000','STRING')
			col_flags2 = 0x02000000
			col_wth_all = YES
			col_wth_kck = NO
			collider = YES
		else:
			if prop.getType() != 'STRING':
				prop.setData('02000000')
				col_flags2 = 0x02000000
				col_wth_all = YES
				col_wth_kck = NO
				collider = YES
			else:
				col_flags2 = int(prop.getData(),16)
				if col_flags2 & CF2_COLLIDE_ALL:
					col_wth_all = YES
					col_wth_kck = NO
					collider = YES
					col_flags2 = col_flags2 & (~CF2_COLLIDE_KCK)
				elif col_flags2 & CF2_COLLIDE_KCK:
					col_wth_all = NO
					col_wth_kck = YES
					collider = NO
					col_flags0 = 0x0000
					cf0 = curobj.getProp('col_flags0')
					cf0.setData('%08x' % col_flags0)
				else:
					col_wth_all = NO
					col_wth_kck = NO
					collider = NO
				prop.setData('%08x' % col_flags2)
		if debug:
			print "Checking 'col_flags3' property."
		try:
			prop = curobj.getProperty('col_flags3')
		except:
			curobj.addProperty('col_flags3','0000','STRING')
		else:
			prop.setData('0000')
		col_flags3 = 0x0000
		if debug:
			print "Checking 'col_flags4' property."
		try:
			prop = curobj.getProperty('col_flags4')
		except:
			curobj.addProperty('col_flags4','00000000','STRING')
			col_flags4 = 0x00000000
			block_cam.val = NO
		else:
			if prop.getType() != 'STRING':
				prop.setData('00000000')
				block_cam.val = NO
				col_flags4 = 0x00000000
			else:
				if col_flags4 & CF4_CAM_AVOID:
					block_cam.val = YES
				else:
					block_cam.val = NO
		if debug:
			print "Checking 'col_flags5' property."
		try:
			prop = curobj.getProperty('col_flags5')
		except:
			if block_cam.val:
				curobj.addProperty('col_flags5','00000004','STRING')
				col_flags5 = 0x00000004
			else:
				curobj.addProperty('col_flags5','00000000','STRING')
			no_fric.val = NO
		else:
			if prop.getType() != 'STRING':
				if block_cam.val:
					prop.setData('00000004')
					col_flags5 = 0x00000004
				else:
					prop.setData('00000000')
					col_flags5 = 0x00000000
				no_fric.val = NO
			else:
				col_flags5 = int(prop.getData(),16)
				if block_cam.val:
					col_flags5 = col_flags5 | CF5_CAM_AVOID
				elif col_flags5 & CF5_CAM_AVOID:
					block_cam.val = YES
					col_flags4 = col_flags4 | CF4_CAM_AVOID
					cf4 = curobj.getProperty('col_flags4')
					cf4.setData('%08x' % col_flags4)
				if col_flags5 & CF5_NO_FRIC:
					no_fric.val = YES
				prop.setData('%08x' % col_flags5)
		if debug:
			print "Checking 'clickfile' property."
		try:
			prop = curobj.getProperty('clickfile')
		except:
			clickable.val = NO
		else:
			if prop.getType() != 'STRING':
				curobj.removeProperty('clickfile')
				clickable.val = NO
			else:
				clickable.val = YES
				script.val = prop.getData()

